
from multiprocessing import Pool
import numpy as np
import io
import sys

try:
    from SQRNdbnseq import DBNToPairs, PairsToDBN, UnAlign, ParseRestraints,\
                           BPMatrix, AnnotateStems, RunSQRNdbnseq, GAPS,\
                           EncodedReactivities, SEPS
    from SQRNalgos import Nussinov, Edmonds, Hungarian
    
except:
    from .SQRNdbnseq import DBNToPairs, PairsToDBN, UnAlign, ParseRestraints,\
                            BPMatrix, AnnotateStems, RunSQRNdbnseq, GAPS,\
                            EncodedReactivities, SEPS
    from .SQRNalgos import Nussinov, Edmonds, Hungarian


def ReAlignDict(shortseq, longseq):
    """ Return the dictionary between unaligned and realigned indices"""

    dct = {}

    i1, i2 = 0, 0

    while i1 < len(shortseq):

        # Shift the longseq index if gap
        if longseq[i2] in GAPS:
            i2 += 1
        else: # otherwise save the pair of matching indices and increment
            dct[i1] = i2
            i1 += 1
            i2 += 1

    return dct


def PrintMatrix(seq, matrix, dbn1='', dbn2=''):
    """Print matrix and sequence into tsv format"""
    print('',*list(seq),sep='\t')

    Pairs1 = DBNToPairs(dbn1) # bps from dbn1 will be framed with []
    Pairs2 = DBNToPairs(dbn2) # bps from dbn2 will be framed with <>

    for i in range(len(seq)):
        print(seq[i], end='\t')
        line = []
        for j in range(len(seq)):
            x = str(matrix[i][j])
            if (i,j) in Pairs1:
                x = '['+x+']'
            if (i,j) in Pairs2:
                x = '<'+x+'>'
            line.append(x)
        print(*line, sep='\t')


def YieldStems(seq, reactivities = None, restraints = None, 
               bpweights = {}, interchainonly = False,
               minlen = 2, minbpscore = 0):
    """Returns a list of stems based on a (possibly aligned) sequence"""

    # turn seq into UPPERCASE & replace T with U
    seq = seq.upper().replace("T", "U")

    # if not restraints - generate a string of dots
    if not restraints:
        restraints = '.'*len(seq)

    # assert the restraints are of matching length
    assert len(seq) == len(restraints)
    
    # Unalign seq, reactivities and restraints strings
    shortseq, shortrest = UnAlign(seq, restraints)

    if reactivities:
        shortreacts = [reactivities[i]
                       for i in range(len(seq))
                       if seq[i] not in GAPS]
    else:
        shortreacts = None
    
    # Parse restraints into unpaired bases (rxs) and base pairs (rbps)
    rbps, rxs, rlefts, rrights = ParseRestraints(shortrest)

    # Build the BP matrices based on the unaligned seq and restraints
    bpboolmatrix, bpscorematrix = BPMatrix(shortseq, bpweights, rxs,
                                           rlefts, rrights,
                                           interchainonly,
                                           reacts = shortreacts)

    # Annotate stems from the matrices and restraint bps
    # using the maxlen definition for stems (diff = 0)
    stems = AnnotateStems(bpboolmatrix, bpscorematrix,
                          rbps, rstems = [],
                          minlen = minlen,
                          minscore = minbpscore,
                          diff = 0)

    # aligned-unaligned indices dictionary
    radict = ReAlignDict(shortseq, seq)

    # Return the stems with re-aligned indices of their base pairs
    return [[[(radict[v], radict[w]) for v, w in stem[0]], stem[-1]]
            for stem in stems]


def mpYieldStems(args):
    """multiprocessing (single-parameter) version of YieldStems"""
    #Unpack args
    seq, reacts, rests, bpweights, interchainonly, minlen, minbpscore = args
    #Call YieldStems
    return YieldStems(seq, reacts, rests,
                      bpweights, interchainonly,
                      minlen, minbpscore)


def MatrixToDBNs(mat, score, depth, verbose = False, sink = sys.stdout):
    """Build dbn structures from the stem-scored matrix
       in a greedy manner"""
    N = mat.shape[0]

    # the bp score threshold
    thr = score*depth

    if True:

        dct   = dict(enumerate(mat.flatten())) # 1-index key: score value
        # Cells sorted in the decreasing order of their scores
        cells = sorted(dct.items(), key = lambda x: x[1], reverse = True)

        # List of [list of base pairs, set of paired positions] lists
        res = [[[],set()],]

        if verbose:
            print(">Conserved base pairs (one by one)", file = sink)

        for cell in cells:
            # Obtain the 2-index base pair from the 1-index
            # E.g., for a square 2x2 matrix the scheme looks like this:
            # 0 -> (0, 0); 1 -> (0, 1); 2 -> (1, 0); 3 -> (1, 1)
            bp = np.unravel_index(cell[0], mat.shape)

            # Ignore all the bps below the threshold
            if cell[1] < thr:
                break

            # Ignore (v, w) bps with v > w as well as
            # the ones forming too short hairpins of <3nt
            if not bp[1]-bp[0] >= 4:
                continue

            # Whether the base pair was added to the existing struct
            # or is in conflict with everything and requires a new struct
            added = False

            for struct in res:
                if bp[0] not in struct[1] and bp[1] not in struct[1]:
                    struct[0].append(bp) # Adding the base pair
                    struct[1].add(bp[0]) # Adding its paired positions
                    struct[1].add(bp[1]) #
                    added = True
                    break
            if not added:
                res.append([[bp,], set(bp)])

            if verbose:
                print(PairsToDBN([bp,], N), round(cell[1], 3), sep='\t', file = sink)

    
    ####################################
    #pairs = sorted([(min(v,w),max(v,w)) for v,w in Nussinov(None, N,
    #                                                        matrix = mat*(mat >= thr))])
    #pairs = sorted([(min(v,w),max(v,w)) for v,w in Edmonds(None, 
    #                                                       matrix = mat*(mat >= thr))])
    #pairs = sorted([(min(v,w),max(v,w)) for v,w in Hungarian(None, N, 
    #                                                       matrix = mat*(mat >= thr))])
    #res = [[pairs,()],]
    ####################################

    # Convert bp-lists to dbns
    dbns = [PairsToDBN(struct[0], N) for struct in res]

    if verbose:
        print(">Conserved base pairs (assembled)", file = sink)
        for dbn in dbns:
            print(dbn, file = sink)
    # Return dbns
    return dbns


def Metrics(ref, pred):
    """Return the calculated metrics values for the two dbns"""
    if not ref:
        return [np.nan]*6

    rb  = set(DBNToPairs(ref))   # reference base pairs
    pb  = set(DBNToPairs(pred))  # predicted base pairs
    TP  = len(pb & rb)
    FP  = len(pb - rb)
    FN  = len(rb - pb)
    PRC = (round(TP / (TP + FP), 3)) if (TP + FP) else 1
    RCL = (round(TP / (TP + FN), 3)) if (TP + FN) else 1
    FSC = (round(2*TP / (2*TP + FP + FN), 3)) if (2*TP + FP + FN) else 1
    return [TP, FP, FN, FSC, PRC, RCL]


def SQRNdbnali(objs, defrests = None, defreacts = None, defref = None,
               bpweights = {}, interchainonly = False,
               minlen = 2, minbpscore = 0,
               threads = 1, verbose = False,
               sink = sys.stdout):
    """Returns a predicted dbn for a set of aligned sequences"""

    # Placeholder LENxLEN matrix of zeros 
    stemmatrix = np.zeros((len(objs[0][1]),len(objs[0][1])))

    # multiprocessing stuff
    with Pool(threads) as pool:

        inputs = [(obj[1],                           # Sequence
                   obj[2],                           # Reactivities
                   defrests if defrests else obj[3], # Restraints
                   bpweights,
                   interchainonly,
                   minlen,
                   minbpscore) for obj in objs]

        # Sum up the stem scores over all sequences
        for stems in pool.imap(mpYieldStems, inputs):
            for stem in stems:
                for v,w in stem[0]:
                    stemmatrix[v, w] += stem[-1]
                    stemmatrix[w, v] += stem[-1]

    # Build the dbns from the assembled stem-scored matrix
    pred_dbns = MatrixToDBNs(stemmatrix, minbpscore, len(objs), verbose, sink = sink)
    # Return the first dbn and the stemmatrix
    return pred_dbns[0], stemmatrix


def mpRunSQRNdbnseq(args):
    """multiprocessing (single-parameter) version of RunSQRNdbnseq;
       here we parallelize over the sequences instead of parallelizing
       over alternative structures within each sequence prediction"""
    obj, paramsetnames, paramsets, threads,\
    rankbydiff, rankby, hardrest, interchainonly,\
    toplim, outplim, conslim, reactformat, verbose, \
    smat, poollim, entropy, algos = args

    evalonly = False

    name, seq, reacts, rests, ref = obj

    # We use a printing buffer so that the output is ordered
    # instead of being mixed up due to parallelization
    with io.StringIO() as buffer:
    
        cons, dbns, nums1, nums2 = RunSQRNdbnseq(name, seq, reacts, rests, ref, paramsetnames,
                                                 paramsets, threads, rankbydiff, rankby,
                                                 hardrest, interchainonly, toplim, outplim,
                                                 conslim, reactformat, evalonly, poollim,
                                                 mp = False, sink = buffer, stemmatrix = smat,
                                                 entropy = entropy, algos = algos)
        return cons, buffer.getvalue()


def Consensus(structs, freqlimit = 0.0, verbose = False, sink = sys.stdout):
    """Builds a Consensus dbn from a list of dbns"""
    bps = {}

    # In how many sequences a base pair
    # has to be present to be added
    # to the consensus
    freqlimit *= len(structs)

    # Count numbers of sequences for each bp
    for struct in structs:
        for bp in DBNToPairs(struct):
            if bp not in bps:
                bps[bp] = 0
            bps[bp] += 1

    resbps = []

    seen = set()

    if verbose:
        print(">Step 2, Populated base pairs", file = sink)

    # Greedily assemble the set of the most frequent non-conflicting bps
    for bp in sorted(bps.keys(), key = lambda x: bps[x], reverse = True):
        if verbose:
            print(PairsToDBN([bp,], len(structs[0])), bps[bp], file = sink)
        if bps[bp] >= freqlimit and bp[0] not in seen and bp[1] not in seen:
            seen.add(bp[0])
            seen.add(bp[1])
            resbps.append(bp)

    # Convert the bps list to dbn and return
    return PairsToDBN(list(set(resbps)), len(structs[0]))


def ReactScore(reacts, seq, dbn):
    """Returns the reactivity score for a predicted structure"""
    if not reacts:
        return 0.5

    # set of paired positions
    paired = set()
    for v,w in DBNToPairs(dbn):
        paired.add(v)
        paired.add(w)

    # how many chain separators we have in the sequence
    sepnum = sum(1 for _ in seq if _ in SEPS)

    # calculate the score as (1 - average-error)
    # where the error for each position is calculated
    # as reactivity for paired positions
    # and (1 - reactivity) for unpaired positions
    # the score ranges from 0.0 (worst) to 1.0 (best)
    reactscore = 1 - sum(reacts[i] if i in paired else 1 - reacts[i]
                         for i in range(len(seq))
                         if seq[i] not in SEPS) / (len(seq) - sepnum)
    return reactscore


def RunSQRNdbnali(objs, defreacts, defrests, defref,
                  levellimit, freqlimit, verbose, step3,
                  paramsetnames, paramsets, threads, rankbydiff, rankby,
                  hardrest, interchainonly, toplim, outplim,
                  conslim, reactformat, poollim, entropy = False,
                  algos = {'G',}, sink = sys.stdout):

    # Alignment length is derived as the length
    # of the first aligned sequence
    N = len(objs[0][1])

    bpweights  = paramsets[0]['bpweights']
    minlen     = paramsets[0]['minlen']
    minbpscore = paramsets[0]['minbpscore']

    if verbose:
        print(">Step 1, Iteration 1", file = sink)

    # Iteration 1    
    pred_dbn, smat = SQRNdbnali(objs, defrests, defreacts, defref, 
                                bpweights, interchainonly,
                                minlen, minbpscore,
                                threads, verbose, sink = sink)

    if verbose:
        print(">Step 1, Iteration 2", file = sink)

    # Iteration 2 - here we use the pred_dbn from the
    # previous iteration instead of defrests
    pred_dbn = SQRNdbnali(objs, pred_dbn, defreacts, defref, 
                          bpweights, interchainonly,
                          minlen, minbpscore,
                          threads, verbose, sink = sink)[0]

    # Truncate pseudoknotted bps of order higher than levellimit for the step1 structure
    step1dbn = PairsToDBN(DBNToPairs(pred_dbn), N,
                          levellimit = levellimit)

    # Normalize the stemmatrix obtained from the first iteration
    smat = smat / np.max(smat) * 5

    if verbose:
        print(">Step 1, Result", file = sink)
        print(step1dbn, file = sink)

    structs = []
    if step3 != '1':
        if verbose:
            print(">Step 2, Individuals", file = sink)
        # Run stemmatrix-weighted single-sequence predictions
        with Pool(threads) as pool:
            inputs = [(obj, paramsetnames, paramsets, threads,
                       rankbydiff, rankby, hardrest, interchainonly,
                       toplim, outplim, conslim, reactformat,
                       verbose, smat, poollim, entropy, algos) for obj in objs]
            for cons, output in pool.imap(mpRunSQRNdbnseq, inputs):
                if verbose:
                    print(output, end = '', file = sink)
                structs.append(cons) # Collect the derived consensus predictions
        # Build the overall consensus from all the single-seq consensus dbns
        step2dbn = Consensus(structs, freqlimit, verbose, sink = sink)
        if verbose:
            print(">Step 2, Consensus", file = sink)
            for lim in range(0, 101, 5):
                print(Consensus(structs, lim / 100), str(lim)+'%', sep='\t', file = sink)
    else:
        step2dbn = '.' * N

    # Truncate pseudoknotted bps of order higher than levellimit for the step2 structure
    step2dbn = PairsToDBN(DBNToPairs(step2dbn), N,
                          levellimit = levellimit)

    if verbose:
        print("=" * N, file = sink) # Separator line 1 if verbose 

    # Print default input lines if any
    if defreacts:
        print(EncodedReactivities(objs[0][1],
                                  defreacts,
                                  reactformat),
              "reactivities", sep='\t', file = sink)
    if defrests:
        print(''.join([defrests[i]
                       if   objs[0][1][i] not in SEPS
                       else objs[0][1][i]
                       for i in range(N)]), "restraints", sep='\t', file = sink)
    if defref:
        print(''.join([defref[i]
                       if   objs[0][1][i] not in SEPS
                       else objs[0][1][i]
                       for i in range(N)]), "reference", sep='\t', file = sink)

    if defreacts or defref or defrests:
        print("_" * N, file = sink) # Separator line 2 if default input lines

        
    print(step1dbn,
          "Step-1" + ('\t'+str(round(ReactScore(defreacts, objs[0][1], step1dbn), 2))) * bool(defreacts),
          "TP={},FP={},FN={},FS={},PR={},RC={}".format(*Metrics(defref, step1dbn)) * bool(defref),
          sep = '\t', file = sink)

    print(step2dbn,
          "Step-2" + "(skipped)" * (step3 == '1') +\
           ('\t'+str(round(ReactScore(defreacts, objs[0][1], step2dbn), 2))) * bool(defreacts) * (step3 != '1'),
          "TP={},FP={},FN={},FS={},PR={},RC={}".format(*Metrics(defref, step2dbn))*bool(defref) * (step3 != '1'),
          sep = '\t', file = sink)

    # Derive the step3dbn based on the step3 param
    if step3 == '1':
        step3dbn = step1dbn
    elif step3 == '2':
        step3dbn = step2dbn
    elif step3 == 'i':
        step3dbn = PairsToDBN(sorted(set(DBNToPairs(step1dbn)) & set(DBNToPairs(step2dbn))), N)
    else: # elif step3 == 'u'
        step1pairs = DBNToPairs(step1dbn)
        seen_pos = set(pos for bp in step1pairs for pos in bp)
        for v, w in DBNToPairs(step2dbn):
            if v not in seen_pos and w not in seen_pos:
                step1pairs.append((v, w))
        step3dbn = PairsToDBN(sorted(step1pairs), N)

    print(step3dbn,
          "Step-3({})".format(step3) +\
           ('\t'+str(round(ReactScore(defreacts, objs[0][1], step3dbn), 2))) * bool(defreacts),
          "TP={},FP={},FN={},FS={},PR={},RC={}".format(*Metrics(defref, step3dbn))*bool(defref),
          sep = '\t', file = sink)


    







    
def BackTrack(begin, end, K, minloop, partial = False):

    queue = {(begin, end), } # start from the last cell
    basepairs = []

    while queue:

        newq = set()

        # for each cell in the queue
        for i,j in queue:
            # if we have (k, j) base pair
            if (i,j) in K:
                k = K[(i,j)]
                # if we may have more base pairs before k  
                if (k - 1) - i > minloop and not partial:
                    newq.add((i, k - 1))
                # if we may have more base pairs inside [k, j]
                if (j - 1) - (k + 1) > minloop:
                    newq.add((k + 1, j - 1))
                # add (k, j) bp
                basepairs.append((k,j))
            # if j is unpaired
            else:
                # if we may have more base pairs inside [i, j - 1]
                if (j - 1) - i > minloop and not partial:
                    newq.add((i, j - 1))
        # update queue
        queue = newq

    return sorted(basepairs)


def Nussinov(stems, N, minloop = 3, matrix = None):

    import numpy as np

    if matrix is None:
        SCORES = {(v,w):-stem[2] for stem in stems for v,w in stem[0]}
    else:
        SCORES = {(v,w):-matrix[v,w]
                  for v in range(N - 1)
                  for w in range(v + 1,N)
                  if matrix[v,w] > 0}

    # score matrix
    D = np.zeros((N,N))
    # base pair dictionary for traceback
    K = {}

    # going diagonale by diagonale,
    # skipping first [minloop] diagonals
    # to avoid too short hairpins
    for h in range(minloop + 1, N):
        # row index
        for i in range(N - h):
            # column index
            j = i + h

            bestk, bestscorek = -1, 10**9
            # searching for the optimal base pair
            for k in range(i, j - minloop):

                if (k,j) in SCORES:
                    scorek = D[i, k - 1] + D[k + 1, j - 1] + SCORES[(k,j)]
                    if scorek < bestscorek:
                        bestk, bestscorek = k, scorek
            # see if no base pair is not better
            if bestscorek <= D[i, j - 1]:

                K[(i, j)] = bestk
                D[i, j]   = bestscorek
            # if it is
            else:
                D[i, j] = D[i, j - 1]

    # traceback
    predictedbps = BackTrack(0, N - 1, K, minloop)
    
    return predictedbps


def Edmonds(stems, power = 1.7, matrix = None):
    
    import networkx as nx

    if matrix is None:
        edges = [(v,w,stem[2]**power) for stem in stems for v,w in stem[0]]
    else:
        edges = [(v,w,matrix[v,w]**power) for v in range(matrix.shape[0] - 1)
                                   for w in range(v + 1, matrix.shape[0])
                                   if matrix[v,w] > 0]

    G = nx.Graph()
    G.add_weighted_edges_from(edges)
    pairs = sorted(nx.max_weight_matching(G))
    return pairs


def Hungarian(stems, N, minloop = 3, power = 1.7, matrix = None):

    import numpy as np
    from scipy.optimize import linear_sum_assignment

    if matrix is None:
        mat = np.zeros((N,N))
        for stem in stems:
            for v,w in stem[0]:
                mat[v,w] = -(stem[2]**power)
                mat[w,v] = -(stem[2]**power)
    else:
        mat = -(matrix**power)

    row_ind, col_ind = linear_sum_assignment(mat)
    sol = {i:j for i,j in zip(row_ind,col_ind)}

    pairs = [(k,sol[k]) for k in sol
             if k < sol[k]-minloop and sol[k] in sol and sol[sol[k]] == k and mat[k,sol[k]] != 0]

    return pairs
    



import os
import sys
import io
from multiprocessing import Pool

try:
    from SQRNdbnseq import RunSQRNdbnseq, ReactDict, SEPS, GAPS
    from SQRNdbnali import RunSQRNdbnali
except:
    from .SQRNdbnseq import RunSQRNdbnseq, ReactDict, SEPS, GAPS
    from .SQRNdbnali import RunSQRNdbnali


def ParseConfig(configfile):
    """Parses the config file"""
    # Set of mandatory parameters
    params = {"algorithms",
              "bpweights",
              "suboptmax",
              "suboptmin",
              "suboptsteps",
              "minlen",
              "minbpscore",
              "minfinscorefactor",
              "distcoef",
              "bracketweight",
              "orderpenalty",
              "loopbonus",
              "maxstemnum"}

    paramsets = []
    names = []
    cnt = 0

    with open(configfile) as file:
        for line in file:
            # Ignore everything after # symbol
            cleanline = line.split('#', 1)[0].strip()
            # If non-empty line
            if cleanline:
                # If paramset name
                if cleanline.startswith('>'):
                    names.append(cleanline[1:])
                    cnt += 1
                    # If this is the first param set
                    if cnt == 1:
                        paramset = {}
                    # Otherwise 
                    else:  
                        paramsets.append(paramset)
                        # Init all the following sets with the first set values
                        paramset = {k:v for k, v in paramsets[0].items()}
                else:
                    key, val = cleanline.split(maxsplit = 1)
                    # bpweights require parsing values like GC=3,AU=2,GU=1
                    if key == "bpweights":
                        paramset[key] = {}
                        for kv in val.split(','):
                            k, v = kv.strip().split('=')
                            paramset[key][k] = float(v)
                    elif key == "algorithms":
                        paramset[key] = set(val.split(','))
                    # all the other params are simply float values
                    else:
                        paramset[key] = float(val)
    # don't forget the last one
    paramsets.append(paramset)

    # Confirm the first param set contains all the params
    if not all([_ in paramsets[0] for _ in params]):
        raise ValueError("Missing some of the parameters in"+\
                         " the first parameter set"+\
                         " of the config file: {}"\
                         .format(', '.join([_ for _ in params
                                            if _ not in paramset])))
    return names, paramsets


def ParseDefaultInput(inputname, inputformat, returndefaults = False, ignore = False):
    """Returns object lists of format [name,sequence,reactivities,restraints,reference]
       or the list [default-reactivities, default-restraints, default reference] if
       returndefaults param is True"""

    warningsT = False
    warningsR = False
    warningsF = False

    def ProcessIndividual(data):
        """Returns a single [seq,reacts,rests,ref] list"""

        nonlocal warningsT, warningsR, warningsF

        while len(data) < len(inputformat):
            data.append(None)

        # split()[0] to allow space-separated comments
        # after the input within the line
        # for sequence, restraints, reference, but not for reactivities
        sequence     = data[q_ind].split()[0]
        reactivities = data[t_ind] if t_ind > 0 else None
        restraints   = data[r_ind].split()[0] if r_ind > 0 and data[r_ind] else None
        reference    = data[f_ind].split()[0] if f_ind > 0 and data[f_ind] else None

        N = len(sequence)

        # Fill features with default values if applicable
        if not reactivities and defT:
            if (len(defT) == N or len(defT.split()) == N):
                reactivities = defT
            elif not warningsT:
                warningsT = True
                if ignore:
                    print("WARNING: some sequences differ in length from the default reactivities line",
                          file=sys.stderr)
                else:
                    raise ValueError("WARNING: some sequences differ in length from the default reactivities line "+\
                                     "[Switch on the iw/ignore parameter to proceed anyway]")
        if not restraints and defR:
            if len(defR) == N:
                restraints   = defR
            elif not warningsR:
                warningsR = True
                if ignore:
                    print("WARNING: some sequences differ in length from the default restraints line",
                          file=sys.stderr)
                else:
                    raise ValueError("WARNING: some sequences differ in length from the default restraints line "+\
                                     "[Switch on the iw/ignore parameter to proceed anyway]")
        if not reference and defF:
            if len(defF) == N:
                reference    = defF
            elif not warningsF:
                warningsF = True
                if ignore:
                    print("WARNING: some sequences differ in length from the default reference line",
                          file=sys.stderr)
                else:
                    raise ValueError("WARNING: some sequences differ in length from the default reference line "+\
                                     "[Switch on the iw/ignore parameter to proceed anyway]")

        # Check reactivities for consistency and resolve them if needed
        try:
            if reactivities:
                if len(reactivities) != len(sequence):
                    reactivities = list(map(float, reactivities.split()))
                else:
                    reactivities = [ReactDict[char] for char in reactivities]

            assert not reactivities or len(reactivities) == len(sequence)
        except:
            raise ValueError('Inappropriate reactivities line for entry "{}":\n {}'\
                             .format(name[1:], reactivities))

        # Assert restraints and reference are of the consistent length
        # or empty line / None
        assert not restraints or len(restraints) == len(sequence),\
               'Inappropriate restraints line for entry "{}":\n {}'\
               .format(name[1:], restraints)
        assert not reference or len(reference) == len(sequence),\
               'Inappropriate reference line for entry "{}":\n {}'\
               .format(name[1:], reference)

        return sequence, reactivities, restraints, reference

    name = None
    defT = None
    defR = None
    defF = None
    data = []

    q_ind = inputformat.index('q')
    t_ind = inputformat.find('t')
    r_ind = inputformat.find('r')
    f_ind = inputformat.find('f')
    
    with open(inputname) as file:
        for line in file:
            if line.startswith('>'):
                # If not the first entry - process the previous one
                if name:
                    yield (name, *ProcessIndividual(data))
                else:
                    # Default TRF lines
                    defdata = data
                    while len(defdata) < len(inputformat) - 1:
                        defdata.append(None)
                    defdata.insert(q_ind, None)
                    defT = defdata[t_ind] if t_ind > 0 else None
                    defR = defdata[r_ind] if r_ind > 0 else None
                    defF = defdata[f_ind] if f_ind > 0 else None

                    if returndefaults:
                        yield (defT, defR, defF)
                        return None
                        
                name = line.strip()
                data = []
            else:
                data.append(line.strip())

    if name:
        yield (name, *ProcessIndividual(data))


def GuessFormat(inp):
    """Identify the input file format: default / fasta / stockholm / clustal"""
    with open(inp) as file:
        line1 = file.readline()

        entry_lines = 0
        seq_lines   = 0

        if line1.startswith('#') and "STOCKHOLM" in line1:
            return "stockholm"
        
        if line1.startswith("CLUSTAL"):
            return "clustal"

        if line1.startswith(">"):
            entry_lines += 1

        for line in file:
            if line.startswith(">"):
                entry_lines += 1
            else:
                acgut = sum(1 for ch in line.upper() if ch in {'A','C','G','U','T'})
                if acgut > len(line) / 2:
                    seq_lines += 1
                if seq_lines > 1000:
                    break

        if seq_lines > entry_lines and entry_lines > 0:
            return "fasta"

    return "default"


def ParseFasta(inp, returndefaults = False):

    if returndefaults:
        yield (None, None, None)
        return None

    name, seq = None, ''

    with open(inp) as file:
        for line in file:
            if line.startswith('>'):
                if name:
                    yield (name, seq, None, None, None)
                name = line.strip()
                seq  = ''
            elif line.strip():
                seq += line.strip()
    yield (name, seq, None, None, None)


def ReadStockholm(stkfile):
    """Parses Stockholm format into three lists and two dicts"""

    seqnames = [] # Sequence names
    seqdict  = {} # Sequence dict with name keys and sequence values
    gcnames  = [] # Structure names
    gcdict   = {} # Structure dict with name keys and structure values
    headers  = [] # Headers list

    try:
        file = open(stkfile)
    except:
        # Non-standard encoding found in some
        # of the Rfam families
        file = open(stkfile, encoding="iso8859-15")

    for line in file:
        if line.startswith('#=GC '): # Structure lines

            linesplit = line.strip().split()
            seq = linesplit[-1]
            name = ' '.join(linesplit[1:-1])

            if name not in gcdict:
                gcnames.append(name)
                gcdict[name] = seq
            else:
                gcdict[name] += seq

        elif line.startswith('#'):
            # Header lines
            headers.append(line)

        elif line.startswith('//'):
            pass
        elif not line.strip():
            pass
        else:
            # Sequence lines
            linesplit = line.strip().split()
            seq = linesplit[-1]
            name = ' '.join(linesplit[:-1])

            if name not in seqdict:
                seqnames.append(name)
                seqdict[name] = seq
            else:
                seqdict[name] += seq

    file.close()

    # Put #=GF lines to the end of the headers
    headers1 = [x for x in headers if not x.startswith("#=GF SQ")]
    headers2 = [x for x in headers if x.startswith("#=GF SQ")]
    headers = headers1 + headers2

    return headers, seqnames, seqdict, gcnames, gcdict


def ParseStockholm(inp, returndefaults = False):
    """Treats SS_cons dbn as default reference"""
    headers, seqnames, seqdict, gcnames, gcdict = ReadStockholm(inp)

    if returndefaults:
        return None, None, gcdict["SS_cons"] if "SS_cons" in gcnames else None

    return [('>'+seqname, seqdict[seqname], None, None,
             gcdict["SS_cons"] if "SS_cons" in gcnames else None)
            for seqname in seqnames]


def ParseClustal(inp, returndefaults = False):

    if returndefaults:
        return None, None, None

    objs = {}
    names = []

    with open(inp) as file:
        for line in file:
            if line.strip() and not line.startswith("CLUSTAL") and not line.startswith(' '):
                name, seq = line.strip().split()
                if name not in objs:
                    names.append(name)
                    objs[name] = ''
                objs[name] += seq

    return [('>'+name, objs[name], None, None, None) for name in names]


def ParseSeq(inputseq, returndefaults):

    if returndefaults:
        return None, None, None
    return [('>inputseq',inputseq,None,None,None),]


def ParseInput(inputseq, inputname, inputformat, returndefaults = False,
               fmt = "unknown", ignore = False):
    """Parser selector"""

    if inputseq:
        return ParseSeq(inputseq, returndefaults), fmt
    
    if fmt == "unknown":
        fmt = GuessFormat(inputname)
        if fmt != "default":
            print("Non-default input file format is recognized: {}".format(fmt.upper()))

    if fmt == "default":
        if returndefaults:
            return next(ParseDefaultInput(inputname, inputformat, returndefaults)), fmt
        return ParseDefaultInput(inputname, inputformat,
                                 returndefaults, ignore = ignore), fmt
    elif fmt == "fasta":
        if returndefaults:
            return next(ParseFasta(inputname, returndefaults)), fmt
        return ParseFasta(inputname, returndefaults), fmt
    elif fmt == "stockholm":
        return ParseStockholm(inputname, returndefaults), fmt
    elif fmt == "clustal":
        return ParseClustal(inputname, returndefaults), fmt

def byseqRunSQRNdbnseq(args):
    """multiprocessing (single-parameter) version of RunSQRNdbnseq;
       here we parallelize over the sequences instead of parallelizing
       over alternative structures within each sequence prediction"""
    name, seq, reacts, restrs, ref, theparamsetnames,\
    theparamsets, threads, rankbydiff, rankby,\
    hardrest, interchainonly, toplim, outplim,\
    conslim, reactformat, evalonly, poollim, entropy, \
    algos, levellimit, priority = args

    # We use a printing buffer so that the output is ordered
    # instead of being mixed up due to parallelization
    with io.StringIO() as buffer:
    
        RunSQRNdbnseq(name, seq, reacts, restrs, ref, theparamsetnames,
                      theparamsets, threads, rankbydiff, rankby,
                      hardrest, interchainonly, toplim, outplim,
                      conslim, reactformat, evalonly, poollim,
                      mp = False, sink = buffer, entropy = entropy,
                      algos = algos, levellimit = levellimit,
                      priority = priority)
        return buffer.getvalue()


def Predict(inputfile = None, fileformat = "unknown", inputseq = None,
            configfile = None, inputformat = "qtrf", maxstemnum = None,
            threads = os.cpu_count(), byseq = False, algorithms = '',
            entropy = False, rankby = "s", evalonly = False, hardrest = False,
            interchainonly = False, toplim = 5, outplim = None, conslim = 1,
            poollim = 1000, reactformat = 3, alignment = False, levellimit = None,
            freqlimit = 0.35, verbose = False, step3 = "u", ignorewarn = False,
            HOME_DIR = None, write_to = None, priority = None,
            i = None, ff = None, c = None, config = None, s = None, seq = None,
            a = None, ali = None, algo = None, algorithm = None, rb = None,
            fl = None, freqlim = None, ll = None, levlim = None, tl = None,
            ol = None, cl = None, pl = None, pr = None, s3 = None, msn = None,
            rf = None, eo = None, hr = None, ico = None, iw = None, ignore = None,
            t = None, bs = None, v = None):
    """
        -----------------------------------------------------------------------
        Prints SQUARNA RNA secondary structure predictions from the given input

        Parameters:
            i / inputfile : string
                Path to the input file.
            ff / fileformat : unknown/fasta/default/stockholm/clustal
                Input file format. For the details of the default SQUARNA input
                format see https://github.com/febos/SQUARNA/blob/main/examples/seq_input.fas.
                "unknown"   - the format will be identified automatically.
                "default"   - default fasta-like format.
                "fasta"     - FASTA format.
                "stockholm" - STOCKHOLM format.
                "clustal"   - CLUSTAL format.
            s / seq / inputseq : string
                RNA sequence input, alternative to inputfile (higher priority).
            c / config / configfile : string
                Path to a config file or a name of a built-in config, 
                see https://github.com/febos/SQUARNA/blob/main/def.conf
                for the format details. 
                In the alignment-based mode, the default config 
                file is ali.conf. In the single-sequence mode the default
                config for sequences under 500nts is def.conf, for sequences
                between 500 and 1000nts - 500.conf, and for sequences over
                1000nts in length - 1000.conf.
                Built-in configs:
                def (def.conf) is recommended by default for RNAs under 500nts.
                alt (alt.conf) is recommended for short pseudoknotted RNAs.
                500 (500.conf) is recommended for RNAs longer 500 nts.
                1000 (1000.conf) is recommended for RNAs longer 1000 nts.
                sk (sk.conf) is recommended with SHAPE data input.
                c=nussinov (nussinov.conf) - Nussinov algorithm config.
                c=hungarian (hungarian.conf) - Hungarian algorithm config.
                c=edmonds (edmonds.conf) - Edmonds algorithm config.
                c=greedy (greedy.conf) - Greedy algorithm config.
            inputformat : string
                The order of the lines in the input file. By default, SQUARNA 
                reads the first line of the entry (among the lines after 
                the ">" line) as the seQuence (q), the second line as the 
                reacTivities (t), the third line as the Restraints (r), 
                the fourth line as the reFerence (f), and all the further lines 
                are ignored. inputformat should be a subset of qtrfx letters 
                in any order, with q being mandatory. All "x" lines will be ignored.
            msn / maxstemnum : int
                Maximum number of stems to predict in each structure. By default,
                maxstemnum is defined in a config file for each parameter set. 
                If specified explicitly it will overwrite the maxstemnum 
                values for all the parameter sets.
            t / threads : int
                Number of CPUs to use.
            bs / byseq : bool
                Parallelize the execution over the input sequences
                in the single-sequence mode. 
                By default, the execution in the single-sequence mode
                is parallelized over the structure pool within each sequence.
                Parallelizing over input sequences is recommended for 
                large input files along with fast configs.
            algo / algorithm / algorithms: string
                The algorithms to be used in single-sequence predictions.
                By default, the algorithms are derived from the config file.
                If the algorithms parameter is specified, it will overwrite the
                algorithms listed in the config file.
                The choice should be a subset of the four algorithms:
                e - Edomnds algorithm [10.6028/jres.069B.013]
                g - Greedy SQUARNA algorithm [10.1101/2023.08.28.555103]
                h - Hungarian algorithm [10.1002/nav.3800020109]
                n - Nussinov algorithm [10.1073/pnas.77.11.6309]
            rb / rankby : string
                How to rank the predicted structures. rankby should be a subset of
                letters r, s, and d in any order (r / s / rs / rd / sd / rsd).
                If both r and s are present, the structures will be ranked according
                to the total_score = structure_score * reactivity_score. If only 
                r is present, the structures will be ranked by the reactivity_score,
                and if only s is present, the structures will be ranked by the 
                structure_score. Independently, if d is present, the mutually 
                divergent structures will be put first.
            eo / evalonly : bool
                Ignored in the alignment mode.
                If specified, no predictions are made and just the reference structure
                scores are returned provided the reference is specified. 
                If non-canonical base pairs are present in the reference structure, 
                they will be considered with 0.0 weight).
            hr / hardrest : bool
                If specified, all the base pairs from the restraints line will be
                forced to be present in the predicted structures. However, it will
                not affect the structure scores, as the forced base pairs won't
                contribute to the structure score unless they were predicted without
                forcing as well.
            ico / interchainonly : bool
                Allow only inter-chain base pairs to be predicted.
            tl / toplim : int
                How many top-N structures will be subject to comparison with the reference.
            ol / outplim : int
                How many top-N structures will be printed into the stdout.
                By default, outplim = toplim.
            cl / conslim : int
                How many top-N structures will be used to derive the predicted structure consensus.
            pl / poollim : int
                Maximum number of structures allowed to populate the current structure pool (if exceeded, no bifurcation will occur anymore).
            rf / reactformat : 3/10/26
                Encoding used to output the reactivities line.
                rf=3:   0.0  <= "_" <   1/3; 
                        1/3  <= "+" <   2/3;
                        2/3  <= "#" <=  1.0;
                rf=10:  0.0  <= "0" <   0.1;
                        ....................
                        0.5  <= "5" <   0.6;
                        ....................
                        0.9  <= "9" <=  1.0;
                rf=26:  0.00 <= "a" <  0.02;
                        0.02 <= "b" <  0.06;
                        ....................
                        0.50 <= "n" <  0.54;
                        ....................
                        0.94 <= "y" <  0.98;
                        0.98 <= "z" <= 1.00.
            a / ali/ alignment : bool
                Run SQUARNA in the alignment-based mode. If specified,
                ali.conf will be used as the config file by default, 
                unless another config file is explicitly specified 
                by the user. The bpweights, minlen, and minbpscore 
                parameters for step-1 will be derived from the first 
                parameter set in the config file.
            ll / levlim / levellimit : int
                Ignored in the single-sequence mode.
                The allowed number of pseudoknot levels. All the base pairs
                of the higher levels will be removed from the structure predicted
                at step-1 and from the structure predicted at step-2. By default, 
                levellimit=3 for short alignments of no more than 500 columns, 
                and levellimit=2 for longer alignments.
            fl / freqlim / freqlimit : 0.0 <= float <= 1.0
                Ignored in the single-sequence mode.
                The percentage of sequences required to contain a base pair,
                in order for it to be added to the predicted consensus structure
                at step-2. The consensus will include all the base pairs present
                in at least "fl" share of the sequences given that the base pair
                is not in conflict (does not share a position) with a more 
                populated base pair.
            v / verbose : bool
                Run SQUARNA in the verbose mode.
                Ignored in the single-sequence mode.
            s3 / step3 : "i"/"u"/"1"/"2"
                Ignored in the single-sequence mode.
                Defines the structure that will be printed at step-3. If step3=1,
                the structure from step-1 will be printed, and the step-2 will
                be skipped completely, meaning the prediction will be super fast.
                If step3=2, the structure from step-2 will be printed. If step3=u,
                the union of base pairs of the two structures will be printed. 
                If step3=i, the intersection of base pairs of the two structures 
                will be printed.
            iw / ignore / ignorewarn : bool
                Ignore warnings.
            HOME_DIR : string
                Path to the folder with built-in configs.
            write_to : IO_object
                Where to write the output. By default, write_to = sys.stdout.
            pr / priority : string
                Comma-separated list of prioritized paramset names. Default: bppN,bppH1,bppH2. 
    """

    # resolve synonyms
    if i != None:
        inputfile = i
    if ff != None:
        fileformat = ff
    if config != None:
        configfile = config
    if c != None:
        configfile = c
    if seq != None:
        inputseq = seq
    if s != None:
        inputseq = s
    if ali != None:
        alignment = ali
    if a != None:
        alignment = a
    if algorithm != None:
        algorithms = algorithm
    if algo != None:
        algorithms = algo
    if rb != None:
        rankby = rb
    if freqlim != None:
        freqlimit = freqlim
    if fl != None:
        freqlimit = fl
    if levlim != None:
        levellimit = levlim
    if ll != None:
        levellimit = ll
    if tl != None:
        toplim = tl
    if ol != None:
        outplim = ol
    if cl != None:
        conslim = cl
    if pl != None:
        poollim = pl
    if pr != None:
        priority = pr
    if s3 != None:
        step3 = s3
    if msn != None:
        maxstemnum = msn
    if rf != None:
        reactformat = rf
    if eo != None:
        evalonly = eo
    if hr != None:
        hardrest = hr
    if ico != None:
        interchainonly = ico
    if ignore != None:
        ignorewarn = ignore
    if iw != None:
        ignorewarn = iw
    if t != None:
        threads = t
    if bs != None:
        byseq = bs
    if v != None:
        verbose = v
    ##################

    if HOME_DIR is None:
        HOME_DIR = os.path.dirname(os.path.abspath(__file__))

    if write_to is None:
        write_to = sys.stdout

    if inputfile != None and not os.path.exists(inputfile) and\
       os.path.exists(os.path.join(HOME_DIR, inputfile)):
        inputfile = os.path.join(HOME_DIR, inputfile)
    
    # Verifying arguments
    assert os.path.exists(str(inputfile)) or inputseq, "Input file does not exist."

    assert fileformat in {'unknown','fasta','default','stockholm','clustal'},\
           "Wrong fileformat, choose one of these: default,fasta,stockholm,clustal"

    if configfile is None:
        configfileset  = False
        configfile     = os.path.join(HOME_DIR, "def.conf")
        configfile500  = os.path.join(HOME_DIR, "500.conf")
        configfile1000 = os.path.join(HOME_DIR, "1000.conf")
        if priority is None:
            priority = set('bppN,bppH1,bppH2'.split(','))
        else:
            priority = set([x for x in priority.split(',') if x])
    else:
        configfileset = True
        if not os.path.exists(configfile):
            if os.path.exists(os.path.join(HOME_DIR, configfile+".conf")):
                configfile = os.path.join(HOME_DIR, configfile+".conf")
            elif os.path.exists(os.path.join(HOME_DIR, configfile)):
                configfile = os.path.join(HOME_DIR, configfile)
        assert os.path.exists(configfile), "Config file does not exist."
        if priority is None:
            priority = set()
        else:
            priority = set([x for x in priority.split(',') if x])

    assert ''.join(sorted(inputformat.replace('x',''))) in {"q","fq","qr","qt", "qrt",
                                                                 "fqr", "fqt", "fqrt"}, \
                   'Inappropriate inputformat value (subset of "fqrtx" with "q" being mandatory): {}'\
                   .format(inputformat)

    if maxstemnum is None:
        maxstemnum = 10**6
        maxstemnumset = False
    else:
        maxstemnumset = True
        try:
            maxstemnum = int(float(maxstemnum))
            assert maxstemnum >= 0
        except:
            raise ValueError("Inappropriate maxstemnum value (non-negative integer): {}"\
                             .format(maxstemnum))
    try:
        threads = int(float(threads))
        threads = max(1, threads)
        threads = min(threads, os.cpu_count())
    except:
        raise ValueError("Inappropriate threads value (integer): {}"\
                         .format(threads))

    try:
        algos = set(algorithms.upper())
        assert algos <= {'E','G','H','N'}
    except:
        raise ValueError('Inappropriate algorithm value (should be subset of "eghn"): {}'\
                         .format(arg.split('=', 1)[1]))

    assert rankby in {"r", "s", "rs", "dr", "ds", "drs"}, \
           'Inappropriate rankby value (r/s/rs/dr/ds/drs): {}'\
           .format(rankby)

    if outplim is None:
        outplim = toplim
        outplimset = False
    else:
        outplimset = True
        try:
            outplim = int(float(outplim))
            assert outplim > 0
            outplimset = True
        except:
            raise ValueError("Inappropriate outplim value (positive integer): {}"\
                             .format(outplim))

    try:
        toplim = int(float(toplim))
        assert toplim > 0
        if not outplimset:
            outplim = toplim
    except:
        raise ValueError("Inappropriate toplim value (positive integer): {}"\
                         .format(toplim))

    try:
        conslim = int(float(conslim))
        assert conslim > 0
    except:
        raise ValueError("Inappropriate conslim value (positive integer): {}"\
                         .format(conslim))
    try:
        poollim = int(float(poollim))
        assert poollim > 0
    except:
        raise ValueError("Inappropriate poollim value (positive integer): {}"\
                         .format(poollim))

    assert int(float(reactformat)) in {3, 10, 26},\
           "Inappropriate reactformat value (3/10/26): {}"\
           .format(reactformat)
    reactformat = int(float(reactformat))

    if not (levellimit is None):
        try:
            levellimit = int(float(levellimit))
        except:
            raise ValueError("Inappropriate levellimit value (integer): {}"\
                             .format(levellimit))

    try:
        freqlimit = float(freqlimit)
        assert 0 <= freqlimit <= 1
    except:
        raise ValueError("Inappropriate freqlimit value (float between 0.0 and 1.0): {}"\
                         .format(freqlimit))

    try:
        step3 = step3.lower()
        assert step3 in {'u', 'i', '1', '2'}
    except:
        raise ValueError("Inappropriate freqlimit value (float between 0.0 and 1.0): {}"\
                         .format(step3))

    # Process rankby
    if "d" in rankby:
        rankbydiff = True # Output diverse structures first
    else:
        rankbydiff = False
    if "r" in rankby and "s" in rankby:
        rankby = (0, 2, 1)
    elif "r" in rankby:
        rankby = (2, 0, 1)
    elif "s"  in rankby:
        rankby = (1, 2, 0)

    # If alignment mode - use ali.conf by default
    if alignment and not configfileset:
        configfile = os.path.join(HOME_DIR, "ali.conf")

    # Parse config
    paramsetnames, paramsets = ParseConfig(configfile)

    # prepare 500.conf & 1000.conf for autoconfig
    if not configfileset:
        paramsetnames500,  paramsets500  = ParseConfig(configfile500)
        paramsetnames1000, paramsets1000 = ParseConfig(configfile1000)

    # Overwrite maxstemnum
    if maxstemnumset:
        for i in range(len(paramsets)):
            paramsets[i]['maxstemnum'] = maxstemnum
        if not configfileset:
            for i in range(len(paramsets500)):
                paramsets500[i]['maxstemnum'] = maxstemnum
            for i in range(len(paramsets1000)):
                paramsets1000[i]['maxstemnum'] = maxstemnum

    # Running single-sequence SQUARNA
    if not alignment:
        # Parallelizing over a structure pool for each sequence
        if not byseq:
            for name, seq, reacts, restrs, ref in ParseInput(inputseq, inputfile, inputformat,
                                                             fmt = fileformat, ignore = ignorewarn)[0]:
                # no autoconfig    
                if configfileset:
                    theparamsetnames, theparamsets = paramsetnames, paramsets
                # apply autoconfig
                else:
                    theparamsetnames, theparamsets = paramsetnames, paramsets
                    if len(seq) >= 500:
                        theparamsetnames, theparamsets = paramsetnames500, paramsets500
                    if len(seq) >= 1000:
                        theparamsetnames, theparamsets = paramsetnames1000, paramsets1000
                
                RunSQRNdbnseq(name, seq, reacts, restrs, ref, theparamsetnames,
                              theparamsets, threads, rankbydiff, rankby,
                              hardrest, interchainonly, toplim, outplim,
                              conslim, reactformat, evalonly, poollim, entropy = entropy,
                              algos = algos, levellimit = levellimit, sink = write_to,
                              priority = priority)
        # Parallelizing over input sequences
        else:
            batchsize = threads*10
            with Pool(threads) as pool:
                inputs_batch = []
                for name, seq, reacts, restrs, ref in ParseInput(inputseq, inputfile, inputformat,
                                                                 fmt = fileformat, ignore = ignorewarn)[0]:
                    # no autoconfig    
                    if configfileset:
                        theparamsetnames, theparamsets = paramsetnames, paramsets
                    # apply autoconfig
                    else:
                        theparamsetnames, theparamsets = paramsetnames, paramsets
                        if len(seq) >= 500:
                            theparamsetnames, theparamsets = paramsetnames500, paramsets500
                        if len(seq) >= 1000:
                            theparamsetnames, theparamsets = paramsetnames1000, paramsets1000
                    # Collecting inputs
                    inputs_batch.append((name, seq, reacts, restrs, ref, theparamsetnames,
                                         theparamsets, threads, rankbydiff, rankby,
                                         hardrest, interchainonly, toplim, outplim,
                                         conslim, reactformat, evalonly, poollim,
                                         entropy, algos, levellimit, priority))
                    # Process a batch once we have batchsize sequences
                    if len(inputs_batch) >= batchsize:
                        for output in pool.imap(byseqRunSQRNdbnseq, inputs_batch):
                            print(output, end = '', file = write_to)
                        inputs_batch = []
                # last batch (if anything left)
                if inputs_batch:
                    for output in pool.imap(byseqRunSQRNdbnseq, inputs_batch):
                        print(output, end = '', file = write_to)
                        

    else: # Running alignment-based SQUARNA

        # Get the processed sequences
        objs, fmt = ParseInput(inputseq, inputfile, inputformat,
                               fmt = fileformat, ignore = ignorewarn)

        # Get the default input lines
        defReactivities, defRestraints, defReference = ParseInput(inputseq, inputfile, inputformat,
                                                                  returndefaults = True,
                                                                  fmt = fmt,
                                                                  ignore = ignorewarn)[0]

        objs = [obj for obj in objs] # Unpack generator (in case of fasta/default format)

        # Length checks
        N = len(objs[0][1])
        assert all(len(obj[1]) == N for obj in objs),\
               'The sequences are not aligned'

        # Check reactivities for consistency and resolve them if needed
        try:
            if defReactivities:
                if len(defReactivities) != N:
                    defReactivities = list(map(float, defReactivities.split()))
                else:
                    defReactivities = [ReactDict[char] for char in defReactivities]

            assert not defReactivities or len(defReactivities) == N
        except:
            raise ValueError('Inappropriate default reactivities line:\n {}'\
                             .format(defReactivities))

        # Assert restraints and reference are of the consistent length
        # or empty line / None
        assert not defRestraints or len(defRestraints) == N,\
               'Inappropriate default restraints line:\n {}'\
               .format(defRestraints)
        assert not defReference or len(defReference) == N,\
               'Inappropriate default reference line:\n {}'\
               .format(defReference)

        # default levellimit
        if levellimit is None:
            levellimit = 3 - int(N > 500)

        # Run the alignment-based predictions
        RunSQRNdbnali(objs, defReactivities, defRestraints, defReference,
                      levellimit, freqlimit, verbose, step3,
                      paramsetnames, paramsets, threads, rankbydiff, rankby,
                      hardrest, interchainonly, toplim, outplim,
                      conslim, reactformat, poollim, entropy = entropy,
                      algos = algos, sink = write_to)


def Main():

    def PrintUsage():
        print()
        print("Usage:")
        print()
        print('pathto/python3 pathto/SQUARNA.py i=inputfile [OPTIONS]')
        print()
        print("For further details read the help message:")
        print()
        print('pathto/python3 pathto/SQUARNA.py --help')
        print()
        exit(1)

    HOME_DIR = os.path.dirname(os.path.abspath(__file__))
    
    args = sys.argv[1:]

    # If no arguments - print the short usage
    if not args:
        PrintUsage()

    # If asking for help message
    if "--help" in args or "-help" in args or "help" in args or\
       "--h" in args or "-h" in args or "h" in args or\
       "--H" in args or "-H" in args or "H" in args:
        with open(os.path.join(HOME_DIR,"README.md")) as helpfile:
            print(helpfile.read())
        exit(0)

    # DEFAULTS
    inputfile      = None
    fileformat     = "unknown"
    inputseq       = None
    configfile     = None

    inputformat = "qtrf"               # Input line order, q=seQuence,t=reacTivities,r=Restraints,f=reFerence

    maxstemnum    = None               # maximum number of stems for each structure

    threads       = os.cpu_count()     # Number of cpus to use
    byseq         = False              # Parallelize by input sequences, not by structure pool

    rankby         = "s"               # Rank by, r / s / rs / dr / ds / drs, r=reactscore,s=structscore,d=rankbydiff
    evalonly       = False             # Just evaluate the reference and do not predict anything
    hardrest       = False             # Force bp-restraints into predicted structures 
    interchainonly = False             # Forbid intra-chain base pairs

    toplim        = 5                  # Top-N to print
    outplim       = None               # Top-N structs used for metrics calculations if reference
    conslim       = 1                  # Top-N structs used for consensus
    poollim       = 100                # Maximum number of structures allowed to populate the current
                                       # structure pool (if exceeded, no bifurcation will occur anymore)

    reactformat   = 3                  # 3 / 10 / 26

    alignment   = False                # Alignment mode
    levellimit  = None                 # Pseudoknot level threshold
    freqlimit   = 0.35                 # The percentage of sequences required to have the base pair
                                       # to include it into the step-2 result (alignment mode)
    verbose     = False                # Print the intermediate output or not
    step3       = "u"                  # i(intersection)/u(union)/1(step1)/2(step2) - what should be the
                                       # step-3 result dbn (alignment mode)

    ignorewarn  = False                # Ignore warnings
    entropy     = False                # Calculate stem matrix entropy
    algorithms  = ""                   # Single-sequence prediction algorithms

    priority    = None                 #Comma-separated list of prioritized paramset names 

    # Allow standard parameter input
    formatted_args = []
    cnt = 0
    while cnt < len(args):
        if args[cnt].lower() in {"-algo", "--algo", "-algorithm", "--algorithm",
                                 "-algos", "--algos", "-algorithms", "--algorithms",
                                 "-i", "--i", "-input", "--input",
                                 "-c", "--c", "-config", "--config",
                                 "-if", "--if", "-inputformat", "--inputformat",
                                 "-rb", "--rb", "-rankby", "--rankby",
                                 "-ff", "--ff", "-fileformat", "--fileformat",
                                 "-fl", "--fl", "-freqlim", "--freqlim",
                                 "-ll", "--ll", "-levlim", "--levlim",
                                 "-tl", "--tl", "-toplim", "--toplim",
                                 "-ol", "--ol", "-outplim", "--outplim",
                                 "-cl", "--cl", "-conslim", "--conslim",
                                 "-pl", "--pl", "-poollim", "--poollim",
                                 "-pr", "--pr", "-priority", "--priority",
                                 "-s3", "--s3", "-step3", "--step3",
                                 "-msn", "--msn", "-maxstemnum", "--maxstemnum",
                                 "-rf", "--rf", "-reactformat", "--reactformat",
                                 "-s", "--s", "-seq", "--seq","-sequence", "--sequence",
                                 "-t", "--t", "-threads", "--threads",}:
            formatted_args.append(args[cnt].lstrip('-')+'='+args[cnt+1])
            cnt += 1
        elif args[cnt].lower() in {"-a", "--a", "-ali", "--ali", "-alignment", "--alignment",
                                   "-bs", "--bs", "-byseq", "--byseq",
                                   "-ent", "--ent", "-entropy", "--entropy",
                                   "-eo", "--eo", "-evalonly", "--evalonly",
                                   "-hr", "--hr", "-hardrest", "--hardrest",
                                   "-iw", "--iw", "-ignore", "--ignore",
                                   "-ico", "--ico", "-interchainonly", "--interchainonly",
                                   "-v", "--v", "-verbose", "--verbose",}:
            formatted_args.append(args[cnt].lstrip('-'))
        else:
            formatted_args.append(args[cnt])
        cnt += 1

    args = formatted_args
    # Parsing arguments
    for arg in args:
        # algorithms
        if arg.lower().startswith("algo=") or\
             arg.lower().startswith("algos=") or\
             arg.lower().startswith("algorithm=") or\
             arg.lower().startswith("algorithms="):
            if not arg.split('=', 1)[1]:
                continue
            algorithms = arg.split('=', 1)[1]
        # inputseq
        if arg.lower().startswith("s=") or\
           arg.lower().startswith("seq=") or\
           arg.lower().startswith("sequence="):
            inputseq = arg.split('=', 1)[1]
        # inputfile
        elif arg.lower().startswith("i=") or\
           arg.lower().startswith("input="):
            inputfile = arg.split('=', 1)[1]
        # fileformat
        elif arg.lower().startswith("ff=") or\
           arg.lower().startswith("fileformat="):
            fileformat = arg.split('=', 1)[1].lower()
        # configfile
        elif arg.lower().startswith("c=") or\
           arg.lower().startswith("config="):
            configfile = arg.split('=', 1)[1]
        # inputformat
        elif arg.lower().startswith("if=") or\
             arg.lower().startswith("inputformat="):
            inputformat = arg.split('=', 1)[1].lower()
        # maxstemnum
        elif arg.lower().startswith("msn=") or\
             arg.lower().startswith("maxstemnum="):
            maxstemnum = arg.split('=', 1)[1]
        # threads
        elif arg.lower().startswith("t=") or\
             arg.lower().startswith("threads="):
            threads = arg.split('=', 1)[1]
        # byseq
        elif arg.lower() in {"bs", "byseq"}:
            byseq = True
        # rankby
        elif arg.lower().startswith("rb=") or\
             arg.lower().startswith("rankby="):
            rankby = ''.join(sorted(arg.split('=', 1)[1].lower()))
        # evalonly
        elif arg.lower() in {"eo", "evalonly"}:
            evalonly = True
        # hardrest
        elif arg.lower() in {"hr", "hardrest"}:
            hardrest = True
        # interchainonly
        elif arg.lower() in {"ico", "interchainonly"}:
            interchainonly = True
        # toplim
        elif arg.lower().startswith("tl=") or\
             arg.lower().startswith("toplim="):
            toplim = arg.split('=', 1)[1]
        # outplim
        elif arg.lower().startswith("ol=") or\
             arg.lower().startswith("outplim="):
            outplim = arg.split('=', 1)[1]
        # conslim
        elif arg.lower().startswith("cl=") or\
             arg.lower().startswith("conslim="):
            conslim = arg.split('=', 1)[1]
        # poollim
        elif arg.lower().startswith("pl=") or\
             arg.lower().startswith("poollim="):
            poollim = arg.split('=', 1)[1]
        # priority
        elif arg.lower().startswith("pr=") or\
             arg.lower().startswith("priority="):
            priority = arg.split('=', 1)[1]
        # reactformat
        elif arg.lower().startswith("rf=") or\
             arg.lower().startswith("reactformat="):
            reactformat = arg.split('=', 1)[1]
        # alignment
        elif arg.lower() in {"a", "ali", "alignment"}:
            alignment = True
        # levellimit
        elif arg.lower().startswith("ll=") or\
             arg.lower().startswith("levlim=") or\
             arg.lower().startswith("levellim=") or\
             arg.lower().startswith("levlimit=") or\
             arg.lower().startswith("levellimit="):
            levellimit = arg.split('=', 1)[1]
        # freqlimit
        elif arg.lower().startswith("fl=") or\
             arg.lower().startswith("freqlim=") or\
             arg.lower().startswith("freqlimit=") or\
             arg.lower().startswith("frequencylim=") or\
             arg.lower().startswith("frequencylimit="):
            freqlimit = arg.split('=', 1)[1]
        # verbose
        elif arg.lower() in {"v", "verbose"}:
            verbose = True
        # ignore
        elif arg.lower() in {"iw", "ignore"}:
            ignorewarn = True
        # entropy
        elif arg.lower() in {"ent", "entropy"}:
            entropy = True
        # step3
        elif arg.lower().startswith("s3=") or\
             arg.lower().startswith("step3="):
            step3 = arg.split('=', 1)[1]
        else:
            if len(args) == 1:
                if os.path.exists(arg):
                    inputfile = arg
                elif sum(arg.lower().count(x) for x in (GAPS | set("acgut"))) > len(arg) / 2:
                    inputseq  = arg
                else:
                    inputfile = arg
            else:
                print("Unrecognized option: {}".format(arg))

    print(inputfile)

    Predict(inputfile, fileformat, inputseq,
            configfile, inputformat, maxstemnum,
            threads, byseq, algorithms, entropy,
            rankby, evalonly, hardrest,
            interchainonly, toplim, outplim, conslim,
            poollim, reactformat, alignment, levellimit,
            freqlimit, verbose, step3, ignorewarn, HOME_DIR,
            None, priority)
        


if __name__ == "__main__":

    Main()


























        
            




import numpy as np
from multiprocessing import Pool
import sys

try:
    from SQRNalgos import Edmonds, Hungarian, Nussinov
except:
    from .SQRNalgos import Edmonds, Hungarian, Nussinov

# Gapped values
GAPS = {'-', '.', '~'}
# Separator values (~chain breaks)
SEPS = {';', '&'}

# Reactivities encoding
ReactDict = {"_" : 0.00,  "+" : 0.50,  "#" : 1.00,
             "0" : 0.05,  "1" : 0.15,  "2" : 0.25,
             "3" : 0.35,  "4" : 0.45,  "5" : 0.55,
             "6" : 0.65,  "7" : 0.75,  "8" : 0.85,
             "9" : 0.95,  "a" : 0.00,  "b" : 0.04,
             "c" : 0.08,  "d" : 0.12,  "e" : 0.16,
             "f" : 0.20,  "g" : 0.24,  "h" : 0.28,
             "i" : 0.32,  "j" : 0.36,  "k" : 0.40,
             "l" : 0.44,  "m" : 0.48,  "n" : 0.52,
             "o" : 0.56,  "p" : 0.60,  "q" : 0.64,
             "r" : 0.68,  "s" : 0.72,  "t" : 0.76,
             "u" : 0.80,  "v" : 0.84,  "w" : 0.88,
             "x" : 0.92,  "y" : 0.96,  "z" : 1.00,
             }


def PrintMatrix(seq, matrix, dbn1='', dbn2=''):
    """Print matrix and sequence into tsv format"""
    print('',*list(seq),sep='\t')

    Pairs1 = DBNToPairs(dbn1) # bps from dbn1 will be framed with []
    Pairs2 = DBNToPairs(dbn2) # bps from dbn2 will be framed with <>

    for i in range(len(seq)):
        print(seq[i], end='\t')
        line = []
        for j in range(len(seq)):
            x = str(matrix[i][j])
            if (i,j) in Pairs1:
                x = '['+x+']'
            if (i,j) in Pairs2:
                x = '<'+x+'>'
            line.append(x)
        print(*line, sep='\t')
        

def EncodedReactivities(seq, reacts, reactformat):
    """Takes a list of floats, returns a string"""

    # Put the reactivities into (0, 1) range
    reacts = [x if 0 <= x <= 1 else 0 if x < 0 else 1 for x in reacts]
    
    if reactformat == 3:
        reactline = ''.join(["_+##"[int(x * 3)]
                             for x in reacts])
    elif reactformat == 10:
        reactline = ''.join(['01234567899'[int(x * 10)]
                             for x in reacts])
    else:
        reactline = ''.join(['abcdefghijklmnopqrstuvwxyz'[int(x * 25) + 0.5]
                             for x in reacts])

    # Introduce the chain separators
    reactline = ''.join([reactline[i] if seq[i] not in SEPS else seq[i]
                         for i in range(len(seq))])
    return reactline


def PairsToDBN(newpairs, length = 0, returnlevels = False, levellimit = -1):
    """Convert a list of base pairs into a dbn string of the given length"""

    # Initialize the dbn string
    dbn = ['.']*length

    # Define "brackets" for 30 pseudoknot levels (and 19 more encoded with cyrillic letters)
    # Higher levels will be simply ignored
    levels = ['()','[]','{}','<>','Aa','Bb','Cc','Dd','Ee','Ff','Gg',
              'Hh','Ii','Jj','Kk','Ll','Mm','Nn','Oo','Pp','Qq','Rr',
              'Ss','Tt','Uu','Vv','Ww','Xx','Yy','Zz',
              'Бб','Гг','Дд','Ёё','Жж','Йй','Лл','Пп',
              'Фф','Цц','Чч','Шш','Щщ','Ьь','Ыы','Ъъ','Ээ','Юю','Яя']

    # groups of non-conflicting base pairs
    groups = [set(),]

    # normalize the pairs (i.e. ensure v < w)
    pairs = set((min(v, w), max(v, w)) for v, w in newpairs)
    
    for pair in sorted(pairs):

        level = 0

        # find the minimum level where the pair is not in conflict
        # with any base pair of that level
        while any(v[0]<=pair[0]<=v[1]<=pair[1] or
                  pair[0]<=v[0]<=pair[1]<=v[1] for v in groups[level]):
            level += 1
            if level == len(groups):
                groups.append(set())
            if level == len(levels):
                levels.append('..')

        # add the pair to the determined level
        groups[level].add(pair)

    # kind of a bubble sort of the base pairs among the levels
    # to maximize the number of base pairs of the lowest levels
    # e.g. to turn (..[[[...)...]]] into [..(((...]...)))
    for times in range(len(groups)-1):
        for i in range(len(groups)-1):

            rest = {v for v in groups[i+1] if any(v[0]<=w[0]<=v[1]<=w[1] or
                                                       w[0]<=v[0]<=w[1]<=v[1]
                                                       for w in groups[i])}
            clean = groups[i+1] - rest

            while rest:

                confjprev = set()
                confiprev = set()

                confj = rest.pop()
                rest.add(confj)
                confj = {confj,}
                confi = {v for v in groups[i] if any(v[0]<=w[0]<=v[1]<=w[1] or
                                                     w[0]<=v[0]<=w[1]<=v[1]
                                                     for w in confj)}

                while confjprev != confj or confiprev != confi:

                    confjprev = confj
                    confiprev = confi

                    confj = {v for v in rest if any(v[0]<=w[0]<=v[1]<=w[1] or
                                                    w[0]<=v[0]<=w[1]<=v[1]
                                                    for w in confi)}
                    confi = {v for v in groups[i] if any(v[0]<=w[0]<=v[1]<=w[1] or
                                                     w[0]<=v[0]<=w[1]<=v[1]
                                                     for w in confj)}

                if len(confi) < len(confj):

                    groups[i]   = confj | (groups[i] - confi)
                    groups[i+1] = confi | (groups[i+1] - confj)

                rest = rest - confj

            if clean:

                groups[i] |= clean
                groups[i+1] -= clean

    if returnlevels:
        levels = {}
        for lev, group in enumerate(groups):
            for bp in group:
                levels[bp] = lev + 1
        return levels

    # remove all levels higher than levellimit (if specified)
    if levellimit >= 0:
        groups = groups[:levellimit]

    # add all the pairs to the dbn string
    # according to their levels  
    for i, group in enumerate(groups):
        for pair in group:
            dbn[pair[0]] = levels[i][0]
            dbn[pair[1]] = levels[i][1]
            
    return ''.join(dbn)


def StemsToDBN(stems, seq):
    """Convert a list of stems (having lists
    of bps as the first elements) into a dbn string"""
    return PairsToDBN([bp for stem in stems for bp in stem[0]], len(seq))


def DBNToPairs(dbn):
    """Convert the dbn string into a sorted list of base pairs"""
    pairs = set()

    # keys == closing brackets, values == matching opening brackets
    closing = {'>':'<',']':'[',')':'(','}':'{','a':'A','b':'B','c':'C','d':'D',
               'e':'E','f':'F','g':'G','h':'H','i':'I','j':'J','k':'K','l':'L',
               'm':'M','n':'N','o':'O','p':'P','q':'Q','r':'R','s':'S','t':'T',
               'u':'U','v':'V','w':'W','x':'X','y':'Y','z':'Z',
               'б':'Б','г':'Г','д':'Д','ё':'Ё','ж':'Ж','й':'Й','л':'Л','п':'П',
               'ф':'Ф','ц':'Ц','ч':'Ч','ш':'Ш','щ':'Щ','ь':'Ь','ы':'Ы','ъ':'Ъ',
               'э':'Э','ю':'Ю','я':'Я',}
    # 30+19 bp stacks for 30+19 allowed pseudoknot levels
    stack = {'<':[],'(':[],'{':[],'[':[],'A':[],'B':[],'C':[],'D':[],'E':[],
             'F':[],'G':[],'H':[],'I':[],'J':[],'K':[],'L':[],'M':[],'N':[],
             'O':[],'P':[],'Q':[],'R':[],'S':[],'T':[],'U':[],'V':[],'W':[],
             'X':[],'Y':[],'Z':[],
             'Б':[],'Г':[],'Д':[],'Ё':[],'Ж':[],'Й':[],'Л':[],'П':[],'Ф':[],
             'Ц':[],'Ч':[],'Ш':[],'Щ':[],'Ь':[],'Ы':[],'Ъ':[],'Э':[],'Ю':[],
             'Я':[],}
              
    for i,v in enumerate(dbn):
        # if we observe an opening bracket
        # then add its index into the matching stack
        if v in stack: 
            stack[v].append(i)
        # else if we observe the closing bracket
        # take the opening index from the matching stack
        # and add the base pair to the pairs set
        elif v in closing:
            # this is to handle closing brackets with no
            # opening partner - they will be ignored
            if stack[closing[v]]:
                pairs.add((stack[closing[v]].pop(), i))

    return sorted(pairs)


def ReAlign(shortdbn, longseq, seqmode=False):
    """ Realign the shortdbn string according to the longseq string
    if seqmode==True the gaps will be hyphens, otherwise - dots"""

    # convert shortdbn into a list
    dbn = [x for x in shortdbn]

    assert len(shortdbn) + sum(longseq.count(GAP) for GAP in GAPS) == len(longseq),\
    "Cannot ReAlign dbn string - wrong number of gaps:\n{}\n{}".format(longseq,shortdbn)

    # initialize the returning string
    newdbn = []

    for x in longseq:
        if x in GAPS: # if gap (hyphen/dot/tilda) -> add gap
            if seqmode:
                newdbn.append('-')
            else:
                newdbn.append('.')
        else: # otherwise -> add the next symbol
            newdbn.append(dbn.pop(0))

    # return the aligned string
    return ''.join(newdbn)


def UnAlign(seq, dbn):
    """Removes gaps from a pair of seq & dbn strings
    hyphens, dots, and tildas are treated as gaps"""

    # First, remove base pairs of gap columns
    # to avoid any shifts
    cleandbn = list(dbn)
    pairs = DBNToPairs(dbn)
    for v,w in pairs:
        if seq[v] in GAPS or seq[w] in GAPS:
            cleandbn[v] = '.'
            cleandbn[w] = '.'

    # Unalign the dbn based on the seq string
    newdbn = ''.join([cleandbn[i] for i in range(len(seq))
                      if seq[i] not in GAPS])
    # Then unalign the seq string itself
    newseq = ''.join([x for x in seq if x not in GAPS])
    # return the unaligned seq & dbn strings
    return newseq, newdbn


def BPMatrix(seq, weights, rxs, rlefts, rrights,
             interchainonly = False, reacts = None,
             bpp_power = 0):
    """Return a matrix with bp-weights in cells"""

    # list of ordinal numbers of chains
    chains = [0 for _ in range(len(seq))]
    if interchainonly:
        curr   =  0
        for i in range(len(seq)):
            if seq[i] in SEPS:
                curr += 1
            else:
                chains[i] = curr

    # Set of different symbols in seq
    bases = set(seq)

    # Dictionary of all possible bp weights
    bps = {}

    # for (x,y) from weights add bps[(x,y)] & bps[(y,x)]
    for bp in weights:
        bps[bp] = weights[bp]
        bps[bp[1]+bp[0]] = weights[bp]

    boolmat = np.zeros((len(seq), len(seq)), dtype = float)

    # Fill the upper triangle of the boolmatrix with ones for bps
    # Ignore the bps that would form a hairpin of len < 3
    # And satisfy the restraints
    for i in range(len(seq) - 1):

        # Allow short "hairpins" between different RNA chains
        inc4 = 4
        for chk in (1, 2):
            if i + chk < len(seq) and seq[i + chk] in SEPS:
                inc4 = chk + 1

        for j in range(i + inc4, len(seq)):
            boolmat[i, j] = int(seq[i] + seq[j] in bps) * \
                            (not interchainonly or chains[i] != chains[j]) * \
                            (i not in rxs and j not in rxs) * \
                            (j not in rlefts) * \
                            (i not in rrights)

    # add 0 values for all possible base pairs
    # not yet in bps dictionary
    for b1 in bases:
        for b2 in bases:
            if b1+b2 not in bps:
                bps[b1+b2] = 0

    # Initialize the bp-score matrix
    scoremat = np.zeros((len(seq), len(seq)), dtype = float)

    # Fill the upper triangle of the matrix with bp weights
    # Ignore the bps that would form a hairpin of len < 3
    for i in range(len(seq) - 1):

        # Allow short "hairpins" between different RNA chains
        inc4 = 4
        for chk in (1, 2):
            if i + chk < len(seq) and seq[i + chk] in SEPS:
                inc4 = chk + 1
        
        for j in range(i + inc4, len(seq)):

            # Weight with reactivities if any
            reactfactor = 1 if reacts is None else (1 - (reacts[i] + reacts[j]) / 2 ) * 2
            if bps[seq[i] + seq[j]] <= 0:
                reactfactor = 1 / max(reactfactor, 0.01)

            scoremat[i, j] = bps[seq[i] + seq[j]] * boolmat[i, j] * reactfactor

    ################
    if bpp_power:
        import RNA
        fc = RNA.fold_compound(''.join([ch if ch not in SEPS else 'N' for ch in seq]))
        fc.pf()
        bppm = np.array(fc.bpp())[1:,1:]
        if np.max(bppm) > 0:
            if bpp_power < 0:
                scoremat += (bppm/np.max(bppm))**(-bpp_power)
            else:
                scoremat *= (bppm/np.max(bppm))**bpp_power
    ################

    return boolmat, scoremat


def ParseRestraints(restraints):
    """ Convert a dbn string into base pairs and unpaired bases"""
    rbps = DBNToPairs(restraints) # Base pairs
    rxs     = {i for i in range(len(restraints)) if restraints[i]=='_'} # Unpaired bases
    rlefts  = {i for i in range(len(restraints)) if restraints[i]=='/'}  # 5'-ends of bps
    rrights = {i for i in range(len(restraints)) if restraints[i]=='\\'} # 3'-ends of bps
    return rbps, rxs, rlefts, rrights


def PreStemsFromDiag(diag):
    "Return longest sequences of consecutive bps"""
    prestems = []
    
    firstbp = -1

    for i, pos in enumerate(diag):

        isbp, val, bp = pos

        # if first bp of the stem
        if isbp and firstbp < 0:
            firstbp = i

        # if first non-bp after the stem
        if not isbp and firstbp >= 0:
            prestems.append(diag[firstbp:i])
            firstbp = -1

    #if the diagonal ends with a stem
    if firstbp >= 0:
        prestems.append(diag[firstbp:])

    return prestems


def StemsFromPreStemsDiffEdge(prestems, diff = 0):
    """Define stems as maximum sub-sequence of consecutive bps with up to
    <diff> trimmed base pairs from either of the edges"""
    
    stems = []
    for prestem in prestems:
        for i in range(diff + 1):
            for j in range(len(prestem) - diff, len(prestem) + 1):
                if j > i:
                    substem = prestem[i:j]
                    bps = [x[2] for x in substem]
                    score = sum(x[1] for x in substem)
                    stems.append([bps, len(bps), score])
    return stems


def StemsFromDiag(diag, diff = 0):
    """Annotate stems at the given diagonal"""
    prestems = PreStemsFromDiag(diag)
    return StemsFromPreStemsDiffEdge(prestems, diff)


def AnnotateStems(bpboolmatrix, bpscorematrix, rbps,
                  rstems, minlen, minscore, diff = 0, span = -1):

    # copy the bpboolmatrix
    matrix = bpboolmatrix.copy()
    # obtain the matrix length
    N = bpboolmatrix.shape[0]
    # initialize result
    stems = []

    # apply bp-restraints - zero out the entire row and column except the cell for each bp
    for v, w in rbps:
        matrix[v, :] *= 0
        matrix[:, v] *= 0
        matrix[w, :] *= 0
        matrix[:, w] *= 0
        matrix[v, w]  = bpboolmatrix[v, w]

    # apply stem-restraints - zero out the entire rows and columns for each stem
    for stem in rstems:
        for v, w in stem[0]:
            matrix[v, :] *= 0
            matrix[:, v] *= 0
            matrix[w, :] *= 0
            matrix[:, w] *= 0

    # start cells of all the diagonals, except the ones producing hairpins of len < 3
    # and ignoring bps with (j - i) > span
    if span < 0 or span >= N:
        diagstarts  = [(0, x)   for x in range(4, N)]
        diagstarts += [(y, N-1) for y in range(1, N-4)]
    else:
        diagstarts = []
        for x in range(4, N):
            if x <= span:
                diagstarts.append((0, x))
            else:
                delta = x - span
                if delta % 2:
                    delta += 1
                delta //= 2
                diagstarts.append((0 + delta, x - delta))
        for y in range(1, N-4):
            if N-1-y <= span:
                diagstarts.append((y, N-1))
            else:
                delta = N-1-y - span
                if delta % 2:
                    delta += 1
                delta //= 2
                diagstarts.append((y + delta, N-1 - delta))

    # for each diagonal select the stems
    # as maximal score sub-arrays
    for x, y in diagstarts:

        diag = []
        i, j = x, y
        
        while i <= j - 1: 
            diag.append([matrix[i, j], bpscorematrix[i, j], (i, j)])
            i += 1
            j -= 1

        for stem in StemsFromDiag(diag, diff):
            if stem[1] >= minlen and stem[2] >= minscore:
                stems.append(stem)
    
    return stems


def PairsToStems(sorted_pairs):

    sp = sorted_pairs

    if len(sp) < 2:
        if not sp:
            return []
        else:
            return [[[sp[0],], 1],]

    stems = [[[sp[0],], 1],]

    for i in range(1, len(sp)):

        if not (sp[i-1][0] + 1 == sp[i][0] and sp[i-1][1] == sp[i][1] + 1):
            stems.append([[], 0])
        stems[-1][0].append(sp[i])
        stems[-1][1] = len(stems[-1][0])

    return stems


def Entropy(bpboolmatrix, bpscorematrix, restbps,
            rstems, minlen, minscore):
    """return mean entropy of the stem matrix"""
    def EntropyRow(row):
        if row.sum():
            probs = [p for p in row / row.sum() if p]
            return sum(-(probs * np.log2(probs)))
        else:
            return 0

    stems = AnnotateStems(bpboolmatrix, bpscorematrix, restbps,
                          rstems, minlen, minscore)

    stemmatrix = np.zeros(bpboolmatrix.shape)
    N = stemmatrix.shape[0]
    for stem in stems:
        score = stem[2]
        for v,w in stem[0]:
            stemmatrix[v,w] = score
            stemmatrix[w,v] = score

    ent = 0
    for i in range(N):
        ent += EntropyRow(stemmatrix[i,:])
    
    return str(round(ent/N,3))


def RunAlgo(bpboolmatrix, bpscorematrix, restbps,
            rstems, minlen, minscore, algo = "E",
            levellimit = 3):
    """Return single-sequence predictions by a specific algorithm"""

    stems = AnnotateStems(bpboolmatrix, bpscorematrix, restbps,
                          rstems, minlen, minscore)

    stemset = []
    pairs   = []

    N = bpboolmatrix.shape[0]

    if algo == "E":
        pairs = Edmonds(stems)

    if algo == "N":
        pairs = Nussinov(stems, N)

    if algo == "H":
        pairs = Hungarian(stems, N)

    # Removing partial stems below thresholds
    stems = PairsToStems(sorted([(min(v,w),max(v,w)) for v,w in pairs]))

    for stem in stems:
        
        score = sum(bpscorematrix[v,w] for v,w in stem[0])
        if score >= minscore and stem[1] >= minlen:
            stemset.append(stem + [score, score, '']) # for consistency with G-algo finstemsets

    # Removing pseudoknots of the order higher than the levellimit
    pairs   = [bp for stem in stemset for bp in stem[0]]
    pairs   = DBNToPairs(PairsToDBN(pairs, N, levellimit = levellimit))
    levels  = PairsToDBN(pairs, N, returnlevels = True)
    stemset = []
    stems = PairsToStems(sorted([(min(v,w),max(v,w)) for v,w in pairs]))

    for index,stem in enumerate(stems):
       
        # & Removing short pseudoknotted stems
        if levels[stem[0][0]] > 1 and stem[1] < 4: # remove short pseudoknotted stems
            continue
        score = sum(bpscorematrix[v,w] for v,w in stem[0])
        if score >= minscore and stem[1] >= minlen:
            stemset.append(stem + [score, score, '']) # for consistency with G-algo finstemsets
            
    return stemset


def IsGNRA(seq):
    """Whether the input sequence is of GNRA pattern"""
    if len(seq) != 4:
        return False
    if seq[0] == 'G' and seq[2] in ('G','A') and seq[3] == 'A':
        return True
    return False


def ScoreStems(seq, stems, rstems,
               reacts, minscore,
               bracketweight, distcoef,
               orderpenalty, loopbonus):
    """Adjust the scores based on the stem distance, pseudoknot level, and
       possibly loop bonus"""

    # short near-symmetric internal loops
    goodloops = {(0, 0), (0, 1), (1, 0),
                 (1, 1), (0, 2), (2, 0),
                 (2, 2), (1, 2), (2, 1),
                 (3, 1), (1, 3),
                 (2, 3), (3, 2),
                 (3, 3), (3, 4), (4, 3), 
                 (4, 4), (4, 2), (2, 4),
                }

    # values == indices of the bp partners or -1 for unpaired bases
    bppartners = [-1 for _ in range(len(seq))]

    # set of restraint-bps
    rbps = set()

    # Fill the bppartners list
    for stem in rstems:
        for bp in stem[0]:
            rbps.add(bp)
            bppartners[bp[0]] = bp[1]
            bppartners[bp[1]] = bp[0]

    # keys = bps, values = their pseudoknot level (starting with 1)
    bplevels = PairsToDBN(rbps, returnlevels = True)

    # calculating adjusted scores for each stem
    for stem in stems:

        descr = ''
        bps = stem[0]

        reactfactor =  1 #(sum((1 - (reacts[bp[0]] + reacts[bp[1]]) / 2 ) for bp in bps) / len(bps) * 2) ** 1.5
        """
        descr = "len={},bps={}".format(stem[1], stem[2]) # for debugging only
        """
        levelset = set() # with base pairs of how many pseudoknot levels the stem is in conflict
        dots = 0 # number of confined unpaired bases
        brackets = 0 # number of bases belonging to the pseudoknotted wings (see URS defs)

        # confined region
        stemstart, stemend = bps[-1]

        block_edges = []

        # if we are inside the sub-ECR - where it will end
        inblockend = -1

        # Whether the stem confines an external loop
        between_chains = False

        for pos in range(stemstart+1, stemend):

            partner = bppartners[pos]

            # if dot and not inside sub-ECR - increment dots
            if partner == -1:

                if pos > inblockend:
                    dots += 1
                # If a separator instead of a dot
                if seq[pos] in SEPS:
                    between_chains = True

            # if the pseudoknotted wing and not inside sub-ECR - increment brackets
            elif partner < stemstart or partner > stemend:

                if pos > inblockend:
                    brackets += 1
                    levelset.add(bplevels[(min(pos, partner),
                                           max(pos, partner))])

            # if sub-ECR face - increase inblockend
            elif pos < partner and partner > inblockend:
                inblockend = partner
                block_edges.append((pos, partner))

        # to prioritize short near-symmetric internal loops inside
        goodloop = False
        diff1 = 0
        if len(block_edges) == 1:
            if (block_edges[0][0] - stemstart - 1,
                stemend - block_edges[0][1] - 1) in goodloops:
                goodloop = True
                diff1 = abs((block_edges[0][0] - stemstart - 1) - (stemend - block_edges[0][1] - 1))
        # to prioritize short near-symmetric internal loops outside
        goodloopout = False
        diff2 = 0
        outerstart, outerend = bps[0]
        vv,ww = outerstart - 1, outerend + 1
        while vv >= 0 and outerstart - vv - 1 < 5 and bppartners[vv] == -1:
            vv -= 1
        while ww < len(seq) and ww - outerend - 1 < 5 and bppartners[ww] == -1:
            ww += 1
        if bppartners[vv] == ww and bppartners[ww] == vv and\
           (outerstart - vv - 1, ww - outerend - 1) in goodloops:
            goodloopout = True
            diff2 = abs((outerstart - vv - 1) - (ww - outerend - 1))

        # Double bonus for symmetric loops, 1.5-bonus for (x-1,x) loops
        # and a single bonus for (x-2,x) loops
        loopfactor = 1 + loopbonus*goodloop*(2-diff1/2) + loopbonus*goodloopout*(2-diff2/2)

        # Bonus for tetraloops
        tetrafactor = 1+0.25*IsGNRA(seq[stemstart+1:stemend])

        # 4 for hairpins, 2 for everything else
        idealdist = 4 if inblockend == -1 else 2

        stemdist = dots + bracketweight*brackets # stem distance

        # Ignore stemdistfactor for external loops
        stemdistfactor = (1 / (1 + abs(stemdist - idealdist)))**distcoef if not between_chains else 1

        order = len(levelset) # number of conflicting pseudoknot levels
        orderfactor = (1 / (1 + order))**orderpenalty

        initscore  = stem[2] # initial bp score
        finalscore = initscore * stemdistfactor * orderfactor * loopfactor * reactfactor * tetrafactor
        """
        descr += ",dt={},br={},sd={},sdf={}".format(dots, brackets,
                                                    round(stemdist,2),
                                                    round(stemdistfactor,2))
        descr += ",or={},orf={}".format(order,
                                        round(orderfactor,2))
        descr += ",lf={},rf={}".format(round(loopfactor,2),
                                       round(reactfactor,2))
        """

        # Forbid lone stems of two bps
        if not goodloop and not goodloopout and len(bps) < 3:
            finalscore = -1
        
        stem.append(finalscore)
        stem.append(descr)

    # remove all stems with finscore < minfinscore
    return [stem for stem in stems if stem[3] >= minscore]


def ChooseStems(allstems, subopt = 1.0):
    """Choose the optimal stems from the stem set"""

    # sort in decreasing order of the adjusted score
    sortedstems = sorted(allstems, key = lambda x: x[3], reverse = True)

    resultstems = sortedstems[:1] # start with the top ranked stem

    # if no stems
    if not resultstems:
        return []

    # the adjusted score of the top ranked stem
    bestscore = resultstems[0][3]

    suboptrange = subopt * bestscore

    # add all stems within the subopt range that are
    # in conflict with all the better stems
    for stem in sortedstems[1:]:

        bps = stem[0]
        finscore = stem[3]

        if finscore < suboptrange:
            return resultstems

        # in conflict == the intersection of their sets
        # of the paired bases is not empty
        if all({p for bp1 in bps
                for p in bp1} & {p for bp2 in betterstem[0]
                                 for p in bp2}
               for betterstem in resultstems):
            resultstems.append(stem)

    return resultstems

 
def OptimalStems(seq, rstems, bpboolmatrix, bpscorematrix,
                 reacts, rbps = set(), 
                 subopt = 1.0, minlen = 2,
                 minbpscore = 6, minfinscore = 0,
                 bracketweight = 1.0, distcoef = 0.1,
                 orderpenalty = 0.0, loopbonus = 0.0,):
    """Return the top stems"""

    # Remove already predicted bps from bp-restraints
    restbps = set(rbps) - {bp for stem in rstems for bp in stem[0]}

    allstems = AnnotateStems(bpboolmatrix, bpscorematrix, restbps,
                             rstems, minlen, minbpscore)

    allstems = ScoreStems(seq, allstems, rstems, reacts,
                          minfinscore,
                          bracketweight, distcoef,
                          orderpenalty, loopbonus,
                          )
    """
    # TEMPORARY PRINTING
    print('##################################################')
    for stem in sorted(allstems, key = lambda x: x[3], reverse = True):
        print(seq)
        print(StemsToDBN(rstems, seq))
        print(StemsToDBN([stem,],seq))
        print(stem[1:])
    """
    
    resultstems = ChooseStems(allstems, subopt)

    """ 
    # TEMPORARY PRINTING
    print('##################################################')
    for stem in resultstems:
        print(seq)
        print(StemsToDBN(rstems, seq))
        print(StemsToDBN([stem,],seq))
        print(stem[1:])
    """ 

    return resultstems


def mpOptimalStems(args):
    """OptimalStems version with a single input parameter for mp.pool.imap"""

    rstems   = args[1] # current stem set
    newstems = OptimalStems(*args)

    return newstems, rstems


def ConsensusStemSet(stemsets):
    """Returns the set of bps present in all the stemsets"""
    
    if not stemsets:
        return set()

    # start with bps of the first stemset
    bps = {bp for stem in stemsets[0] for bp in stem[0]}

    # and intersect with all the other stemsets
    for stemset in stemsets[1:]:
        bps &= {bp for stem in stemset for bp in stem[0]}

    return bps


def ScoreStruct(seq, stemset, reacts):
    """ Return the overall structure score based on the stemset"""
    bpscores = {"GU": -0.5,
                "AU":  1.5,
                "GC":  4.0,
               }
    for bp in sorted(bpscores.keys()): # CG for GC, etc.
        bpscores[bp[::-1]] = bpscores[bp]

    power = 1.7 # we will sum up these powers of the stem scores
    thescore = 0
    
    paired = set()

    for stem in stemset:
        
        bpsum = 0
        for v,w in stem[0]:
            bpsum += (bpscores[seq[v]+seq[w]] if seq[v]+seq[w] in bpscores else 0.0)
            paired.add(v)
            paired.add(w)
        
        if bpsum > 0:
            thescore += bpsum**power 

    # seq length without the separator positions
    sepnum = sum(1 for _ in seq if _ in SEPS)

    # reactscore - sum of "errors" for each position
    # treating paired positions as predicted to be 0
    # and unpaired positions as predicted to be 1
    # then for each paired position the error is the
    # reactivity, and for unpaired = 1-reactivity
    reactscore = 1 - sum(reacts[i] if i in paired else 1 - reacts[i]
                         for i in range(len(seq))
                         if seq[i] not in SEPS) / (len(seq) - sepnum)

    # Return three scores - totalscore, structscore, reactscore
    return round(thescore * reactscore, 3), round(thescore, 3), round(reactscore, 3)


def RankStructs(stemsets, rankbydiff = False, rankby = (0, 2, 1),
                priority = set()):##############################
    """Rank the predicted structures"""

    # First sort according to the rankby order of scores
    finstemsets = sorted(stemsets,
                         key = lambda x: [x[1][rb] for rb in rankby],
                         reverse = True)

    ########################
    finstemsets = [stemset for stemset in finstemsets if priority & set(stemset[2])] +\
                  [stemset for stemset in finstemsets if not (priority & set(stemset[2]))]
    ########################

    # if not rankbydiff - that is all
    if not rankbydiff or len(finstemsets) < 3:
        return finstemsets

    # Otherwise - iteratively select the furthest from the
    # already selected structures breaking ties by the scores
    allbps = set()

    # Compile the entire set of observed bps among all
    # the structures
    for stemset in finstemsets:
        bps = {bp for stem in stemset[0] for bp in stem[0]}
        stemset.append(bps)
        allbps |= bps

    seenbps = {bp for bp in finstemsets[0][-1]}
    curind  = 1

    # While the pre-selected subset of structures
    # does not include all the observed bps
    # or until the last but one structure
    while seenbps != allbps and curind < len(finstemsets) - 1:

        finstemsets = finstemsets[:curind] +\
                      sorted(finstemsets[curind:],
                             key = lambda x: (len(x[-1] - seenbps),
                                              [x[1][rb] for rb in rankby]),
                             reverse = True)
        seenbps |= finstemsets[curind][-1]
        curind += 1

    # Sort the remaining tail according to the rankby
    # order of the scores
    finstemsets = finstemsets[:curind] +\
                  sorted(finstemsets[curind:],
                         key = lambda x: [x[1][rb] for rb in rankby],
                         reverse = True)
    
    # Remove the appended bps sets and return
    return [x[:-1] for x in finstemsets]  


def ReferenceScores(seq, ref, reacts):
    """Returns the three scores for the reference structure"""

    # unalign everything
    if not reacts:
        reacts = [0.5 for i in range(len(seq))]
    reacts = [reacts[i] for i in range(len(seq)) if seq[i] not in GAPS]
    seq, ref = UnAlign(seq, ref)

    pairs   = DBNToPairs(ref)
    stemset = PairsToStems(sorted(pairs))

    return ScoreStruct(seq, stemset, reacts)

   
def SQRNdbnseq(seq, reacts = None, restraints = None, dbn = None,
               paramsets = [], conslim = 1, toplim = 5,
               hardrest = False, rankbydiff = False,
               rankby = (0, 2, 1), interchainonly = False,
               threads = 1, mp = True, stemmatrix = None, poollim = 1000,
               entropy = False, algos = set(), levellimit = None,
               priority = set()):##############
    """seq == sequence (possibly with gaps or any other non-ACGU symbols
    reacts == list of reactivities (values from 0 to 1)
    restraints == string in dbn format; x symbols are forced to be unpaired
    dbn == known secondary structure in dbn format
    conslim == how many alternative structures are used to derive the consensus
    toplim == how many top ranked structures are used to measure the performance
    hardrest == force restraint base pairs to be present in the predicted dbns or not
    rankbydiff == Rank the most different structures on top
    rankby == indices of the scores used for rankings
    interchainonly = allow only bps between different chains
    threads == number of CPUs to use
    mp == whether to use multiprocessing at all
    stemmatrix == if specified the bpscorematrix will be weighted with the stemmatrix
                  that was derived from the sequence alignment
    poollim == limit for bifurcations - stop keeping sub-optimal stems if we have more than poollim structures
    entropy == return mean entropy of the stem matrix under the first paramset instead of predictions
    algos   == single-sequence prediction algorithms to use
    levellimit == pseudoknot order limit (non-pseudoknotted structure is level = 1)
    SQRNdbnseq returns a list of alternative predicted secondary structures in dbn format"""

    assert set(rankby) == {0, 1, 2} and len(rankby) == 3, "Invalid ranking indices"

    # turn seq into UPPERCASE & replace T with U
    seq = seq.upper().replace("T", "U")

    # if not restraints - generate a string of dots
    if not restraints:
        restraints = '.'*len(seq)

    assert len(seq) == len(restraints), "Invalid restraints given"

    # default reacts == list of 0.5 values
    if not reacts:
        reacts = [0.5 for i in range(len(seq))]

    assert len(reacts) == len(seq), "Invalid reactivities given"

    # if we need to decode the reactivities from a string
    if type(reacts) == str:
        reacts = [ReactDict[ch] for ch in reacts]
    
    # Unalign seq, dbn, reacts, and restraints strings
    shortseq, shortrest   = UnAlign(seq, restraints)
    shortreacts = [reacts[i] for i in range(len(seq)) if seq[i] not in GAPS]

    if dbn:
        assert len(seq) == len(dbn)
        shortseq, shortdbn  = UnAlign(seq, dbn)

    # Unalign stemmatrix
    if stemmatrix is not None:
        gapinds = [i for i in range(len(seq)) if seq[i] in GAPS]
        shortsmat = np.delete(stemmatrix, gapinds, 0)
        shortsmat = np.delete(shortsmat, gapinds, 1)
    
    # Parse restraints into unpaired bases (rxs) and base pairs (rbps)
    rbps, rxs, rlefts, rrights = ParseRestraints(shortrest)

    # final stem sets among all the parameter sets
    finfinstemsets  = []
    seen_structures = {} # dict of bp-sets (to avoid repeats in finfinstemsets) bps:{psi1,psi2,...}

    if levellimit is None:
        levellimit = 3 - int(len(shortseq) > 500)

    defalgos = {a for a in algos}

    for psi, paramset in enumerate(paramsets):

        bpweights         = paramset["bpweights"]
        bpp               = paramset["bpp"]
        suboptmax         = paramset["suboptmax"]
        suboptmin         = paramset["suboptmin"]
        suboptsteps       = paramset["suboptsteps"]
        minlen            = paramset["minlen"]
        minbpscore        = paramset["minbpscore"]
        minfinscorefactor = paramset["minfinscorefactor"]
        bracketweight     = paramset["bracketweight"]
        distcoef          = paramset["distcoef"]
        orderpenalty      = paramset["orderpenalty"]
        maxstemnum        = paramset["maxstemnum"]
        loopbonus         = paramset["loopbonus"]

        # use the algorithms from the config file if not specified by the user
        if not defalgos:
            algos = paramset['algorithms']

        # starting subopt value
        cursubopt = suboptmin
        # increment for the subopt value
        suboptinc = (suboptmax - suboptmin)/suboptsteps

        minfinscore = minbpscore * minfinscorefactor

        # Generate initial bp-matrix (with bp weights in cells)
        bpboolmatrix, bpscorematrix = BPMatrix(shortseq, bpweights, rxs,
                                               rlefts, rrights,
                                               interchainonly,
                                               reacts = shortreacts,
                                               bpp_power = bpp)

        #Weight the bpscorematrix with the alignment-derived scores
        if stemmatrix is not None:
            bpscorematrix = bpscorematrix * shortsmat

        if entropy:
            return Entropy(bpboolmatrix, bpscorematrix, rbps,
                           [], minlen, minbpscore)

        # List of finalized stem lists
        finstemsets = []

        for algo in algos: # Run Nissinov/Hungarian/Edmonds
            if algo == "G": # The Greedy algorithm is performed below
                continue
            stemset = RunAlgo(bpboolmatrix, bpscorematrix, rbps,
                              [], minlen, minbpscore, algo = algo,
                              levellimit = levellimit)
            finstemsets.append(stemset)

        if "G" in algos:

            # List of lists of stems (each stem list is a currently predicted secondary structure
            curstemsets = [[],]

            # Starting with a single empty structure
            cursize = len(curstemsets)

            if mp: # multiprocessing version
                with Pool(threads) as pool:

                    # while list of intermediate stem lists is not empty
                    while curstemsets:
                        # Each time we diverge - cursubopt is increased by suboptinc
                        if len(curstemsets) > cursize:
                            cursize = len(curstemsets)
                            if cursubopt < suboptmax:
                                cursubopt += suboptinc

                        # filtering by len(stems) == maxstemnum
                        newcurstemsets = []
                        for stems in curstemsets:
                            if len(stems) == maxstemnum:
                                finstemsets.append(stems)
                            else:
                                newcurstemsets.append(stems)
                        curstemsets = newcurstemsets

                        # new iteration
                        newcurstemsets = []

                        inputs = ((shortseq, stems,
                                   bpboolmatrix, bpscorematrix,
                                   shortreacts, rbps, 
                                   cursubopt, minlen, minbpscore, 
                                   minfinscore, bracketweight,
                                   distcoef, orderpenalty,
                                   loopbonus,
                                   ) for stems in curstemsets)

                        # new optimal stems based on the current stem list 
                        for newstems, stems in pool.imap(mpOptimalStems, inputs):

                            # append new intermediate stem lists
                            if newstems:
                                stopper =  1 if cursize >= poollim else len(newstems)
                                for newstem in newstems[:stopper]:
                                    newcurstemsets.append(stems + [newstem,])
                            # if no newstems returned - the stem list is considered final
                            else:
                                finstemsets.append(stems)

                        # update the current stem lists
                        curstemsets = newcurstemsets

            else: # non-multiprocessing version
                # while list of intermediate stem lists is not empty
                while curstemsets:

                    # Each time we diverge - cursubopt is increased by suboptinc
                    if len(curstemsets) > cursize:
                        cursize = len(curstemsets)
                        if cursubopt < suboptmax:
                            cursubopt += suboptinc

                    # filtering by len(stems) == maxstemnum
                    newcurstemsets = []
                    for stems in curstemsets:
                        if len(stems) == maxstemnum:
                            finstemsets.append(stems)
                        else:
                            newcurstemsets.append(stems)
                    curstemsets = newcurstemsets

                    # new iteration
                    newcurstemsets = []

                    for stems in curstemsets:

                        # new optimal stems based on the current stem list
                        newstems = OptimalStems(shortseq, stems, bpboolmatrix,
                                                bpscorematrix, shortreacts, rbps,
                                                cursubopt, minlen, minbpscore, 
                                                minfinscore, bracketweight,
                                                distcoef, orderpenalty,
                                                loopbonus)

                        # append new intermediate stem lists
                        if newstems:
                            stopper =  1 if cursize >= poollim else len(newstems)
                            for newstem in newstems[:stopper]:
                                newcurstemsets.append(stems + [newstem,])
                        # if no newstems returned - the stem list is considered final
                        else:
                            finstemsets.append(stems)

                    # update the current stem lists
                    curstemsets = newcurstemsets

        for finstemset in finstemsets:

            bpsset = tuple(sorted([bp for stem in finstemset for bp in stem[0]]))

            if bpsset not in seen_structures:
                # append [stemset, structscore, paramsetind]
                finfinstemsets.append([finstemset,
                                       ScoreStruct(shortseq, finstemset, shortreacts),
                                       psi])
                seen_structures[bpsset] = {psi, }
            else:
                seen_structures[bpsset].add(psi)

    # list of dbn strings
    dbns = []

    #Change the first paramset-id to a full set of paramset-ids
    for i in range(len(finfinstemsets)):
        bpsset = tuple(sorted([bp for stem in finfinstemsets[i][0] for bp in stem[0]]))
        finfinstemsets[i][2] = sorted(seen_structures[bpsset])

    # sort the final stem lists in descreasing order of their total bp-score by default
    # and if rankbydiff - prioritize the most diverged structures
    finstemsets = RankStructs(finfinstemsets, rankbydiff, rankby, priority = priority)#############################

    forcedbps = {(v,w) for v,w in rbps
                 if shortseq[v]+shortseq[w] in bpweights or
                    shortseq[w]+shortseq[v] in bpweights} if hardrest else set()

    # convert all final stem lists into dbn strings
    # and not forget about non-predicted bps from restraints
    for stems, structscore, paramsetind in finstemsets:
        dbns.append(PairsToDBN({bp for stem in stems for bp in stem[0]} | forcedbps,
                    len(shortseq)))

    consbps = ConsensusStemSet([xx[0] for xx in finstemsets[:conslim]]) | forcedbps # Consensus of the Top-ranked

    # ReAlign the dbn strings accoring to seq
    dbns = [ReAlign(x, seq) for x in dbns]
    cons = ReAlign(PairsToDBN(consbps, len(shortseq)), seq)

    # Introducing chain separators into the predicted structures
    dbns = [''.join([_[i] if seq[i] not in SEPS else seq[i]
                     for i in range(len(seq))]) for _ in dbns]
    cons = ''.join([cons[i] if seq[i] not in SEPS else seq[i]
                    for i in range(len(seq))])

    # if input dbn is known - calculate the quality metrics
    if dbn:
        knownbps = set(DBNToPairs(shortdbn))

        constp = len(consbps & knownbps)
        consfp = len(consbps - knownbps)
        consfn = len(knownbps - consbps)

        consprc = (round(constp / (constp + consfp), 3)) if (constp + consfp) else 1
        consrcl = (round(constp / (constp + consfn), 3)) if (constp + consfn) else 1
        consfsc = (round(2*constp / (2*constp + consfp + consfn), 3)) if (2*constp + consfp + consfn) else 1

        consresult = [constp, consfp, consfn, consfsc, consprc, consrcl]

        bestfsc = -1
        result = []

        for rank, stemset in enumerate(finstemsets):

            setbps = {bp for stem in stemset[0] for bp in stem[0]} | forcedbps

            tp = len(setbps & knownbps) # Correctly predicted bps
            fp = len(setbps - knownbps) # Wrongly predicted bps
            fn = len(knownbps - setbps) # Missed bps

            prc = (round(tp / (tp + fp), 3)) if (tp + fp) else 1
            rcl = (round(tp / (tp + fn), 3)) if (tp + fn) else 1
            fsc = (round(2*tp / (2*tp + fp + fn), 3)) if (2*tp + fp + fn) else 1

            if fsc > bestfsc:
                bestfsc = fsc 
                result  = [tp, fp, fn, fsc, prc, rcl, rank + 1]

            # TopN only (N == toplim)
            if rank + 1 >= toplim:
                break

        return cons, [(dbns[jj],*finstemsets[jj][1:]) for jj in range(len(dbns))], consresult, result
    return cons, [(dbns[jj],*finstemsets[jj][1:]) for jj in range(len(dbns))], [np.nan]*6, [np.nan]*7


def RunSQRNdbnseq(name, sequence, reactivities, restraints,
                  reference, paramsetnames,
                  paramsets, threads, rankbydiff, rankby,
                  hardrest, interchainonly, toplim, outplim,
                  conslim, reactformat, evalonly, poollim = 1000,
                  mp = True, sink = sys.stdout, stemmatrix = None,
                  entropy = False, algos = {'G',}, levellimit = None,
                  priority = None,):
    """Main-like function;
       sink param is standard system output by default,
       we need it to use the buffer in alignment-mode parallelizations"""

    print(name, file = sink)

    ###########
    if priority:
        priority = {i for i in range(len(paramsetnames)) if paramsetnames[i] in priority}
    else:
        priority = set()
    ###########

    if entropy:
        entropy_val = SQRNdbnseq(sequence, reactivities, restraints, reference,
                                 paramsets, conslim, toplim, hardrest,
                                 rankbydiff, rankby, interchainonly, threads, mp, stemmatrix,
                                 poollim, entropy = True, algos = algos)
        print('\t'.join([sequence,"entropy:",entropy_val]), file = sink)
    else:
        print(sequence, file = sink)

    # Printing everything observed in the input
    if reactivities:
        print(EncodedReactivities(sequence,
                                  reactivities,
                                  reactformat),
              "reactivities", sep = '\t', file = sink)
    if restraints:
        print(''.join([restraints[i]
                       if sequence[i] not in SEPS
                       else sequence[i]
                       for i in range(len(sequence))]),
              "restraints", sep = '\t', file = sink)
    if reference:
        print(''.join([reference[i]
                       if sequence[i] not in SEPS
                       else sequence[i]
                       for i in range(len(sequence))]),
              "reference",
              *ReferenceScores(sequence, reference, reactivities),
              sep = '\t', file = sink)

    # Separator line 1
    print('_'*len(sequence), file = sink)

    if evalonly:
        return None, None, None, None

    # Run prediction
    prediction = SQRNdbnseq(sequence, reactivities, restraints, reference,
                            paramsets, conslim, toplim, hardrest,
                            rankbydiff, rankby, interchainonly, threads, mp, stemmatrix,
                            poollim, algos = algos, levellimit = levellimit, priority = priority)###########
    
    # Unpack the results
    consensus, predicted_structures, consensus_metrics, topN_metrics = prediction

    # Printing consensus
    # along with its metrics if reference is present
    if reference:
        print(consensus,
              "top-{}_consensus".format(conslim),
              "TP={},FP={},FN={},FS={},PR={},RC={}".format(*consensus_metrics),
              sep = '\t', file = sink)
    else:
        print(consensus,
              "top-{}_consensus".format(conslim), sep = '\t', file = sink)

    # Separator line 2
    print('='*len(sequence), file = sink)

    # Printing up to outplim predicted structures
    # along with their scores and
    # metrics of the best one if reference is present
    for i, pred in enumerate(predicted_structures[:outplim]):
        
        struct, scores, paramsetinds = pred
        totalscore, structscore, reactscore = scores

        if reference and i + 1 == topN_metrics[-1]:
            print(struct, "#{}".format(i+1), totalscore,
                  structscore, reactscore,
                  ','.join([paramsetnames[paramsetind]
                            for paramsetind in paramsetinds]),
                  "TP={},FP={},FN={},FS={},PR={},RC={},RK={}".format(*topN_metrics),
                  sep='\t', file = sink)
        else:
            print(struct, "#{}".format(i+1), totalscore,
                  structscore, reactscore,
                  ','.join([paramsetnames[paramsetind]
                            for paramsetind in paramsetinds]),
                  sep='\t', file = sink)

    return consensus, predicted_structures, consensus_metrics, topN_metrics


